# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozphab.conduit import (
    conduit,
)
from mozphab.exceptions import (
    Error,
)
from mozphab.logger import (
    logger,
)
from .submit import (
    add_submit_arguments,
    submit,
)

UPLIFT_REPOSITORY_TAG_SLUG = "uplift"


def map_train_arg_to_repo(train: str) -> dict:
    """Attempt to map the value of `--train` to a Phabricator repo."""
    repo = conduit.get_repository(train)
    if not repo:
        raise Error(f"No repo named {train} on Phabricator!")

    uplift_repos = conduit.get_repositories_with_tag(UPLIFT_REPOSITORY_TAG_SLUG)
    if not uplift_repos:
        raise Error("No uplift repos found on Phabricator! Please file a bug.")

    uplift_repo_callsigns = {uplift["fields"]["callsign"] for uplift in uplift_repos}

    if repo["fields"]["callsign"] not in uplift_repo_callsigns:
        raise Error(f"Repo {train} is not an uplift repository.")

    return repo


def list_trains():
    """List all uplift repositories on Phabricator.

    Queries the Phabricator repository search endpoint for all repositories
    with the `uplift` tag set on them, and displays their callsigns as a list
    of valid values for `--train`.
    """
    repositories = conduit.get_repositories_with_tag(UPLIFT_REPOSITORY_TAG_SLUG)
    if not repositories:
        raise Error("Found no repos with `uplift` tag on Phabricator.")

    logger.info("Available trains for uplift:")
    for repository in repositories:
        logger.info(f"   - {repository['fields']['callsign']}")


def uplift(repo, args):
    if args.list_trains:
        return list_trains()

    if not args.train:
        raise Error("Missing `--train` argument!")

    phab_repo = map_train_arg_to_repo(args.train)

    # When using `moz-phab uplift`, we perform a standard `submit` but with all
    # the in-repo configurations for submit destination overwritten to submit
    # to the uplift target repository instead.
    repo.call_sign = args.train
    repo._phab_repo = phab_repo

    # Run the usual submit comment with our patched arg values.
    submit(repo, args)

    logger.warning(
        "\nPlease navigate to the tip-most commit and complete the uplift "
        "request form."
    )


def add_parser(parser):
    uplift_parser = parser.add_parser(
        "uplift",
        help="Submit uplift request commit(s) to Phabricator.",
        description=(
            "MozPhab will create a new revision to request patches be uplifted "
            "to stable release trains."
        ),
    )

    # Add all the same arguments as `moz-phab submit`
    add_submit_arguments(uplift_parser)

    # Add uplift-specific arguments
    uplift_parser.add_argument(
        "--train",
        help=(
            "Indicate Phabricator callsign of the release train this stack "
            "should be uplifted to."
        ),
    )
    uplift_parser.add_argument(
        "--list-trains",
        help="List all possible values for `--train`.",
        action="store_true",
    )
    uplift_parser.add_argument(
        "--no-rebase",
        help=(
            "Send the specified range of commits as-is for uplift - do not attempt "
            "to rebase."
        ),
        action="store_true",
    )
    uplift_parser.set_defaults(func=uplift, needs_repo=True)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import builtins
import datetime
import pytest
import subprocess
import unittest

from pathlib import Path
from unittest import mock

from mozphab.commands import submit
from mozphab import (
    conduit,
    detect_repository,
    environment,
    exceptions,
    helpers,
    mozphab,
    simplecache,
    subprocess_wrapper,
)


class Helpers(unittest.TestCase):
    @mock.patch("builtins.open")
    @mock.patch("mozphab.helpers.json")
    def test_read_json_field(self, m_json, m_open):
        m_open.side_effect = FileNotFoundError
        self.assertEqual(None, helpers.read_json_field(["nofile"], ["not existing"]))

        m_open.side_effect = NotADirectoryError
        with self.assertRaises(NotADirectoryError):
            helpers.read_json_field(["nofile"], ["not existing"])

        m_open.side_effect = ValueError()
        self.assertEqual(None, helpers.read_json_field(["nofile"], ["not existing"]))

        m_open.side_effect = None
        m_json.load.return_value = dict(a="value A", b=3)
        self.assertEqual(None, helpers.read_json_field(["filename"], ["not existing"]))
        self.assertEqual("value A", helpers.read_json_field(["filename"], ["a"]))

        m_json.load.side_effect = (
            dict(a="value A", b=3),
            dict(b="value B", c=dict(a="value CA")),
        )
        self.assertEqual(3, helpers.read_json_field(["file_a", "file_b"], ["b"]))
        m_json.load.side_effect = (
            dict(b="value B", c=dict(a="value CA")),
            dict(a="value A", b=3),
        )
        self.assertEqual(
            "value B", helpers.read_json_field(["file_b", "file_a"], ["b"])
        )
        m_json.load.side_effect = (
            dict(a="value A", b=3),
            dict(b="value B", c=dict(a="value CA")),
        )
        self.assertEqual(
            "value CA", helpers.read_json_field(["file_a", "file_b"], ["c", "a"])
        )

    @mock.patch.object(builtins, "input")
    @mock.patch("mozphab.mozphab.sys")
    def test_prompt(self, m_sys, m_input):
        input_response = None

        def _input(_):
            return input_response

        m_input.side_effect = _input

        # Default
        input_response = ""
        self.assertEqual("AAA", helpers.prompt("", ["AAA", "BBB"]))

        # Escape
        m_sys.exit.side_effect = SystemExit()
        with self.assertRaises(SystemExit):
            input_response = chr(27)
            helpers.prompt("", ["AAA"])

        with self.assertRaises(SystemExit):
            input_response = chr(27)
            helpers.prompt("")

        input_response = "aaa"
        self.assertEqual("AAA", helpers.prompt("", ["AAA", "BBB"]))
        input_response = "a"
        self.assertEqual("AAA", helpers.prompt("", ["AAA", "BBB"]))
        input_response = "b"
        self.assertEqual("BBB", helpers.prompt("", ["AAA", "BBB"]))
        input_response = "abc"
        self.assertEqual("abc", helpers.prompt(""))

    @mock.patch("mozphab.detect_repository.probe_repo")
    def test_repo_from_args(self, m_probe):
        # TODO test walking the path
        repo = None

        def probe_repo(_):
            return repo

        m_probe.side_effect = probe_repo

        class Args:
            def __init__(self, path=None):
                self.path = path

        with self.assertRaises(exceptions.Error):
            detect_repository.repo_from_args(Args(path="some path"))

        repo = mock.MagicMock()
        args = Args(path="some path")
        self.assertEqual(repo, mozphab.repo_from_args(args))
        repo.set_args.assert_called_once_with(args)

    def test_strip_differential_revision_from_commit_body(self):
        self.assertEqual("", helpers.strip_differential_revision("\n\n"))
        self.assertEqual(
            "",
            helpers.strip_differential_revision(
                "\nDifferential Revision: http://phabricator.test/D123"
            ),
        )
        self.assertEqual(
            "",
            helpers.strip_differential_revision(
                "Differential Revision: http://phabricator.test/D123"
            ),
        )
        self.assertEqual(
            "title",
            helpers.strip_differential_revision(
                "title\nDifferential Revision: http://phabricator.test/D123"
            ),
        )
        self.assertEqual(
            "title",
            helpers.strip_differential_revision(
                "title\n\nDifferential Revision: http://phabricator.test/D123"
            ),
        )
        self.assertEqual(
            "title\n\nsummary",
            helpers.strip_differential_revision(
                "title\n\n"
                "summary\n\n"
                "Differential Revision: http://phabricator.test/D123"
            ),
        )

    def test_amend_commit_message_body_with_new_revision_url(self):
        self.assertEqual(
            "\nDifferential Revision: http://phabricator.test/D123",
            submit.amend_revision_url("", "http://phabricator.test/D123"),
        )
        self.assertEqual(
            "title\n\nDifferential Revision: http://phabricator.test/D123",
            submit.amend_revision_url("title", "http://phabricator.test/D123"),
        )
        self.assertEqual(
            "\nDifferential Revision: http://phabricator.test/D123",
            submit.amend_revision_url(
                "\nDifferential Revision: http://phabricator.test/D999",
                "http://phabricator.test/D123",
            ),
        )

    @mock.patch("mozphab.helpers.os.access")
    @mock.patch("mozphab.helpers.os.path")
    @mock.patch("mozphab.helpers.which")
    def test_which_path(self, m_which, m_os_path, m_os_access):
        m_os_path.exists.side_effect = (True, False)
        m_os_access.return_value = True
        m_os_path.isdir.return_value = False

        path = "x"
        self.assertEqual(path, helpers.which_path(path))
        m_which.assert_not_called()
        helpers.which_path(path)
        m_which.assert_called_once_with(path)


@mock.patch("mozphab.conduit.ConduitAPI.call")
def test_valid_reviewers_in_phabricator_returns_no_errors(call_conduit):
    # See https://phabricator.services.mozilla.com/api/user.search
    call_conduit.side_effect = (
        # user.query
        [{"userName": "alice", "phid": "PHID-USER-1"}],
        # project.search
        {
            "data": [{"fields": {"slug": "user-group"}, "phid": "PHID-PROJ-1"}],
            "maps": {
                "slugMap": {
                    "alias1": {"slug": "name1", "projectPHID": "PHID-PROJ-2"},
                    "#alias2": {"slug": "name2", "projectPHID": "PHID-PROJ-3"},
                }
            },
        },
    )
    reviewers = dict(granted=[], request=["alice", "#user-group", "#alias1", "#alias2"])
    assert [] == conduit.conduit.check_for_invalid_reviewers(reviewers)


@mock.patch("mozphab.conduit.ConduitAPI.call")
def test_disabled_reviewers(call_conduit):
    reviewers = dict(granted=[], request=["alice", "goober"])
    call_conduit.side_effect = (
        # user.query
        [
            dict(userName="alice", phid="PHID-USER-1"),
            dict(userName="goober", phid="PHID-USER-2", roles=["disabled"]),
        ],
        # project.search
        {"data": [], "maps": {"slugMap": {}}},
    )
    expected_errors = [dict(name="goober", disabled=True)]
    assert expected_errors == conduit.conduit.check_for_invalid_reviewers(reviewers)


@mock.patch("mozphab.conduit.ConduitAPI.call")
def test_non_existent_reviewers_or_groups_generates_error_list(call_conduit):
    ts = 1543622400
    ts_str = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H:%M")
    reviewers = dict(
        granted=[],
        request=[
            "alice",
            "goober",
            "goozer",
            "#user-group",
            "#goo-group",
            "#gon-group",
        ],
    )
    call_conduit.side_effect = (
        # user.query
        [
            dict(userName="alice", phid="PHID-USER-1"),
            dict(
                userName="goober",
                phid="PHID-USER-2",
                currentStatus="away",
                currentStatusUntil=ts,
            ),
        ],
        # project.search
        {
            "data": [{"fields": {"slug": "user-group"}, "phid": "PHID-PROJ-1"}],
            "maps": {"slugMap": {}},
        },
    )
    expected_errors = [
        dict(name="#gon-group"),
        dict(name="#goo-group"),
        dict(name="goober", until=ts_str),
        dict(name="goozer"),
    ]

    errors = conduit.conduit.check_for_invalid_reviewers(reviewers)
    errors.sort(key=lambda k: k["name"])
    assert expected_errors == errors


@mock.patch("mozphab.conduit.ConduitAPI.call")
def test_reviewer_case_sensitivity(call_conduit):
    reviewers = dict(granted=[], request=["Alice", "#uSeR-gRoUp"])
    call_conduit.side_effect = (
        # See https://phabricator.services.mozilla.com/conduit/method/user.query/
        [dict(userName="alice", phid="PHID-USER-1")],
        # See https://phabricator.services.mozilla.com/conduit/method/project.search/
        {
            "data": [{"fields": {"slug": "user-group"}, "phid": "PHID-PROJ-1"}],
            "maps": {"slugMap": {}},
        },
    )
    assert [] == conduit.conduit.check_for_invalid_reviewers(reviewers)


def test_get_users_no_users():
    conduit = mozphab.conduit
    assert [] == conduit.get_users([])


@mock.patch("mozphab.conduit.ConduitAPI.call")
def test_get_users_with_user(m_conduit):
    conduit = mozphab.conduit

    user = {"userName": "alice", "phid": "PHID-USER-1"}
    m_conduit.return_value = [user]
    assert [user] == conduit.get_users(["alice"])
    m_conduit.assert_called_once()

    conduit.get_users(["alice"])
    m_conduit.assert_called_once()

    simplecache.cache.reset()
    m_conduit.reset_mock()
    m_conduit.return_value = []
    assert [] == conduit.get_users(["alice"])
    m_conduit.assert_called_once()


def test_simple_cache():
    cache = simplecache.SimpleCache()
    assert cache.get("nothing") is None

    cache.set("something", 123)
    assert cache.get("something") == 123

    assert cache.get("SoMeThInG") == 123

    cache.set("something", "foo")
    assert cache.get("something") == "foo"

    cache.delete("something")
    assert cache.get("something") is None


@mock.patch("subprocess.check_output")
@mock.patch("mozphab.subprocess_wrapper.logger")
def test_check_output(m_logger, m_check_output):
    m_check_output.side_effect = subprocess.CalledProcessError(
        cmd=["some", "cmd"], returncode=2, output="output msg", stderr="stderr msg"
    )
    with pytest.raises(exceptions.CommandError) as e:
        subprocess_wrapper.check_output(["command"])

    assert e.value.status == 2
    assert str(e.value).startswith("command 'command'")
    assert mock.call("stderr msg") in m_logger.debug.call_args_list
    assert mock.call("output msg") in m_logger.debug.call_args_list

    m_check_output.side_effect = ("response \nline \n",)
    assert ["response ", "line"] == subprocess_wrapper.check_output(["command"])

    m_check_output.side_effect = ("response \nline \n",)
    assert ["response ", "line "] == subprocess_wrapper.check_output(
        ["command"], strip=False
    )

    m_check_output.side_effect = ("response \nline \n",)
    assert "response \nline" == subprocess_wrapper.check_output(
        ["command"], split=False
    )


def test_git_find_repo(git_repo_path):
    path = str(git_repo_path)
    assert path == detect_repository.find_repo_root(path)
    subdir = git_repo_path / "test_dir"
    subdir.mkdir()
    assert path == detect_repository.find_repo_root(str(subdir))


def test_hg_find_repo(hg_repo_path):
    path = str(hg_repo_path)
    assert path == detect_repository.find_repo_root(path)


def test_fail_find_repo():
    path = "/non/existing/path"
    assert detect_repository.find_repo_root(path) is None


@mock.patch("mozphab.detect_repository.Mercurial")
@mock.patch("mozphab.detect_repository.Git")
def test_probe_repo(m_git, m_hg):
    m_hg.return_value = "HG"

    assert "HG" == detect_repository.probe_repo("path")

    m_hg.side_effect = ValueError
    m_git.return_value = "GIT"
    assert "GIT" == detect_repository.probe_repo("path")

    m_git.side_effect = ValueError
    assert detect_repository.probe_repo("path") is None


@mock.patch("mozphab.detect_repository.probe_repo")
def test_repo_from_args(m_probe):
    # TODO test walking the path
    repo = None

    def probe_repo(path):
        return repo

    m_probe.side_effect = probe_repo

    class Args:
        def __init__(self, path=None):
            self.path = path

    with pytest.raises(exceptions.Error):
        detect_repository.repo_from_args(Args(path="some path"))

    repo = mock.MagicMock()
    args = Args(path="some path")
    assert repo == detect_repository.repo_from_args(args)
    repo.set_args.assert_called_once_with(args)


def test_parse_config():
    res = helpers.parse_config(
        ["key=value 1", "key2 = value2 ", "key3=", "key4=one=two=three"]
    )
    assert res == dict(key="value 1", key2="value2", key3="", key4="one=two=three")


def test_parse_config_key_only():
    assert helpers.parse_config(["key"]) == dict()


def test_parse_config_with_filter():
    def _filter(name, value):
        if name != "out":
            return True

    res = helpers.parse_config(["imin=I'm in", "out=not here"], _filter)
    assert res == dict(imin="I'm in")


@mock.patch("os.path.expanduser")
@mock.patch("os.path.join")
@mock.patch("os.path.isfile")
@mock.patch("os.stat")
@mock.patch("os.chmod")
@mock.patch("os.getenv")
def test_get_arcrc_path(m_getenv, m_chmod, m_stat, m_isfile, m_join, m_expand):
    arcrc = helpers.get_arcrc_path

    m_expand.return_value = "arcrc file"
    m_isfile.return_value = False
    arcrc()
    m_chmod.assert_not_called()

    class Stat:
        st_mode = 0o100600

    stat = Stat()
    m_stat.return_value = stat
    m_isfile.return_value = True

    m_chmod.reset_mock()
    simplecache.cache.reset()
    arcrc()
    m_chmod.assert_not_called()

    m_chmod.reset_mock()
    m_getenv.reset_mock()
    m_join.reset_mock()
    m_getenv.side_effect = ("/app_data",)
    stat.st_mode = 0o100640
    simplecache.cache.reset()
    arcrc()
    if environment.IS_WINDOWS:
        m_getenv.assert_called_once_with("APPDATA", "")
        m_join.assert_called_once_with("/app_data", ".arcrc")
    else:
        m_chmod.assert_called_once_with("arcrc file", 0o600)


def test_short_node():
    assert (
        helpers.short_node("b016b6080ff9fa6d9ac459950e24bdcdaa939be0") == "b016b6080ff9"
    )
    assert (
        helpers.short_node("this-is-not-a-sha-this-is-not-a-sha-test")
        == "this-is-not-a-sha-this-is-not-a-sha-test"
    )
    assert helpers.short_node("b016b6080ff9") == "b016b6080ff9"
    assert helpers.short_node("b016b60") == "b016b60"
    assert helpers.short_node("mozilla-central") == "mozilla-central"


def test_temporary_file_unicode():
    message = "ćwikła"
    with helpers.temporary_file(message) as fname:
        with Path(fname).open(encoding="utf-8") as f:
            assert f.readline() == message


@pytest.mark.parametrize("prefix", ("+", "-", " "))
@pytest.mark.parametrize("linesep", ("\n", "\r\n"))
class TestCreateHunkLines:
    def test_create_hunk_lines_empty_body_do_not_check_eof(self, prefix, linesep):
        """Expect no lines when provided a completely empty file."""
        lines, eof_has_no_new_line = helpers.create_hunk_lines(
            body="", prefix=prefix, check_eof=False
        )
        assert lines == []
        assert eof_has_no_new_line is None

    def test_create_hunk_lines_empty_body_check_eof(self, prefix, linesep):
        """Expect "no newline message" when provided a completely empty file."""
        lines, eof_had_no_newline = helpers.create_hunk_lines(
            body="", prefix=prefix, check_eof=True
        )
        if prefix != "+":
            assert lines == [
                f"\\ No newline at end of file\n",
            ]
            assert eof_had_no_newline
        else:
            assert lines == []
            assert eof_had_no_newline is None

    def test_create_hunk_lines_empty_line_do_not_check_eof(self, prefix, linesep):
        """Expect a prefixed line to be returned and the EOF check to not proceed.

        NOTE: Though this functionality is currently implemented, it does not match
        what Mercurial does with new, empty files that do not have any content. In those
        cases, Mercurial does not provide any hunks at all regardless if a newline
        characters exists or does not exist at the end of the file.
        """
        lines, eof_has_no_new_line = helpers.create_hunk_lines(
            body=f"{linesep}", prefix=prefix, check_eof=False
        )
        assert lines == [f"{prefix}{linesep}"]
        assert eof_has_no_new_line is None

    def test_create_hunk_lines_empty_line_check_eof(self, prefix, linesep):
        """Expect a prefixed line to be returned, and the EOF check to pass.

        NOTE: Though this functionality is currently implemented, it does not match
        what Mercurial does with new, empty files that do not have any content. In those
        cases, Mercurial does not provide any hunks at all regardless if a newline
        characters exists or does not exist at the end of the file.
        """
        lines, eof_had_no_newline = helpers.create_hunk_lines(
            body=f"{linesep}", prefix=prefix, check_eof=True
        )
        assert lines == [f"{prefix}{linesep}"]
        assert not eof_had_no_newline

    def test_create_hunk_lines_no_newline_eof_do_not_check_eof(self, prefix, linesep):
        """Expect a list of two lines, the last line having no line separator."""
        lines, eof_had_no_newline = helpers.create_hunk_lines(
            body=f"hello{linesep}world", prefix=prefix, check_eof=False
        )
        assert lines == [
            f"{prefix}hello{linesep}",
            f"{prefix}world",
        ]
        assert eof_had_no_newline is None

    def test_create_hunk_lines_no_newline_eof_check_eof(self, prefix, linesep):
        """Expect a list of three lines, all having line separators.

        NOTE: The last message is appended as an extra line, to match with what
        Mercurial outputs when showing diffs.
        """
        lines, eof_had_no_newline = helpers.create_hunk_lines(
            body=f"hello{linesep}world", prefix=prefix, check_eof=True
        )
        assert lines == [
            f"{prefix}hello{linesep}",
            f"{prefix}world\n",
            f"\\ No newline at end of file\n",
        ]
        assert eof_had_no_newline

    def test_create_hunk_lines_disallowed_prefix(self, prefix, linesep):
        """Expect an exception to be raised when a disallowed prefix is provided"""
        with pytest.raises(ValueError):
            helpers.create_hunk_lines("hello world", "*")


class TestSplitLines:
    """Tests the `helpers.split_lines` method.

    The tests here are meant to illustrate the functionality of `helpers.split_lines`
    method, as there are a few special cases (most notably when an input string begins
    or ends with a newline character.) The tests cover POSIX and DOS style line
    terminators, as well as when an input contains a combination of both."""

    def test_empty_body(self):
        """Expect a single, empty string entry when given an empty body."""
        body = ""
        expected = [""]
        actual = helpers.split_lines(body)
        assert actual == expected

    def test_single_line_no_eof_linesep(self):
        """Expect a single string entry when provided with a string with no newlines."""
        body = "line1"
        expected = ["line1"]
        actual = helpers.split_lines(body)
        assert actual == expected

    def test_two_lines_no_eof_linesep_posix(self):
        """Expect a list of lines and newline characters when given multiple lines."""
        body = "line1\nline2"
        expected = ["line1", "\n", "line2"]
        actual = helpers.split_lines(body)
        assert actual == expected

    def test_two_lines_eof_linesep_posix(self):
        """Expect last entry to be empty string when input is newline terminated."""
        body = "line1\nline2\n"
        expected = ["line1", "\n", "line2", "\n", ""]
        actual = helpers.split_lines(body)
        assert actual == expected

    def test_two_lines_no_eof_linesep_dos(self):
        """Test two lines without EOF CRLF."""
        body = "line1\r\nline2"
        expected = ["line1", "\r\n", "line2"]
        actual = helpers.split_lines(body)
        assert actual == expected

    def test_two_lines_eof_linesep_dos(self):
        """Test two lines with CRLF line terminators. Last entry should be empty."""
        body = "line1\r\nline2\r\n"
        expected = ["line1", "\r\n", "line2", "\r\n", ""]
        actual = helpers.split_lines(body)
        assert actual == expected

    def test_three_lines_mix_linesep_no_eof_linesep(self):
        """Test a mix of dos and posix line separators. Result should include both."""
        body = "line1\r\nline2\nline3"
        expected = ["line1", "\r\n", "line2", "\n", "line3"]
        actual = helpers.split_lines(body)
        assert actual == expected

    def test_three_lines_mix_linesep_eof_linesep(self):
        """Test a mix of line separators with EOF CRLF. Last entry should be empty."""
        body = "line1\r\nline2\nline3\r\n"
        expected = ["line1", "\r\n", "line2", "\n", "line3", "\r\n", ""]
        actual = helpers.split_lines(body)
        assert actual == expected

    def test_three_newline_characters(self):
        """Test three empty lines. First and last entries should be empty."""
        body = "\n\n\n"
        expected = ["", "\n", "", "\n", "", "\n", ""]
        actual = helpers.split_lines(body)
        assert actual == expected

        body = "\n\r\n\n"
        expected = ["", "\n", "", "\r\n", "", "\n", ""]
        actual = helpers.split_lines(body)
        assert actual == expected

        body = "\r\n\r\n\r\n"
        expected = ["", "\r\n", "", "\r\n", "", "\r\n", ""]
        actual = helpers.split_lines(body)
        assert actual == expected

    def test_one_line_followed_by_newline_characters(self):
        """Test one line followed by newlines. Last entry should be empty."""
        body = "line1\n\n\n"
        expected = ["line1", "\n", "", "\n", "", "\n", ""]
        actual = helpers.split_lines(body)
        assert actual == expected

    def test_form_feed_character_mixed_in(self):
        """Test handling of form feed character."""
        body = "line1\nline2\fstill on line2\nline3\n"
        expected = ["line1", "\n", "line2\fstill on line2", "\n", "line3", "\n", ""]
        actual = helpers.split_lines(body)
        assert actual == expected


class TestJoinLineseps:
    def test_empty_list(self):
        """Empty list of lines should return empty list."""
        lines = []
        expected = []
        actual = helpers.join_lineseps(lines)
        assert expected == actual

    def test_one_line(self):
        """A list with a single entry should return the same entry back."""
        lines = ["line1"]
        expected = ["line1"]
        actual = helpers.join_lineseps(lines)
        assert expected == actual

    def test_one_line_with_newline(self):
        """A list of two entries should return one entry with the joined string."""
        lines = ["line1", "\n"]
        expected = ["line1\n"]
        actual = helpers.join_lineseps(lines)
        assert expected == actual

    def test_two_lines_no_eof_newline(self):
        """A list of three entries should return the first two joined and the last."""
        lines = ["line1", "\n", "line2"]
        expected = ["line1\n", "line2"]
        actual = helpers.join_lineseps(lines)
        assert expected == actual

    def test_two_lines_with_eof_newline(self):
        """A list of four entries should return a list of two joined strings."""
        lines = ["line1", "\n", "line2", "\n"]
        expected = ["line1\n", "line2\n"]
        actual = helpers.join_lineseps(lines)
        assert expected == actual

    def test_two_empty_lines(self):
        """A list of four entries should return a list of two joined strings.

        This test is basically the same test as above, but added to illustrate the usage
        of this method to join empty lines.
        """
        lines = ["", "\n", "", "\n"]
        expected = ["\n", "\n"]
        actual = helpers.join_lineseps(lines)
        assert expected == actual

    def test_form_feed_character(self):
        """Form feed character should be treated like any other character."""
        lines = ["line1", "\n", "line2\fstill on line2"]
        expected = ["line1\n", "line2\fstill on line2"]
        actual = helpers.join_lineseps(lines)
        assert expected == actual


def test_augment_commits_from_body():
    # The actual parsing is tested in test_commit_parsing; this tests that the commits
    # structure is updated correctly.

    commits = [
        {
            "title": "Bug 1 - test r?reviewer",
            "body": "Differential Revision: https://example.com/D101",
        },
        {
            "title": "WIP: Bug 2 - blah r=blocker!",
            "body": "Differential Revision: https://example.com/D102",
        },
    ]
    helpers.augment_commits_from_body(commits)

    assert commits[0]["rev-id"] == "101"
    assert commits[0]["bug-id"] == "1"
    assert commits[0]["bug-id-orig"] == "1"
    assert commits[0]["reviewers"]["request"] == ["reviewer"]
    assert commits[0]["reviewers"]["granted"] == []
    assert commits[0]["title-preview"] == "Bug 1 - test r?reviewer"
    assert not commits[0]["wip"]

    assert commits[1]["rev-id"] == "102"
    assert commits[1]["bug-id"] == "2"
    assert commits[1]["bug-id-orig"] == "2"
    assert commits[1]["reviewers"]["request"] == []
    assert commits[1]["reviewers"]["granted"] == ["blocker!"]
    assert commits[1]["title-preview"] == "WIP: Bug 2 - blah r=blocker!"
    assert commits[1]["wip"]


def test_move_drev_to_original():
    # Ensure the arguments are returned as-is when `rev_id` is `None`.
    assert helpers.move_drev_to_original("blah", None) == (
        "blah",
        None,
    ), "Passing `None` for `rev_id` should return the arguments."

    # Ensure `Differential Revision` is moved to `Original` and `rev_id` is wiped.
    commit_message = (
        "bug 1: title r?reviewer\n"
        "\n"
        "Differential Revision: http://phabricator.test/D1"
    )
    expected = (
        "bug 1: title r?reviewer\n" "\n" "Original Revision: http://phabricator.test/D1"
    )
    message, rev_id = helpers.move_drev_to_original(commit_message, 1)
    assert (
        message == expected
    ), "`Differential Revision` not re-written to `Original Revision` on uplift."
    assert rev_id is None, "`rev_id` not returned as `None` for new uplift."

    # Ensure `Original` and `Differential` in commit message is recognized as update.
    commit_message = (
        "bug 1: title r?reviewer\n"
        "\n"
        "Original Revision: http://phabricator.test/D1"
        "\n"
        "Differential Revision: http://phabricator.test/D2"
    )
    message, rev_id = helpers.move_drev_to_original(commit_message, 2)
    assert (
        message == commit_message
    ), "Commit message should not have changed when updating an uplift."
    assert rev_id == 2, "`rev_id` should not have changed when updating an uplift."


def test_validate_email():
    invalid_email = "Test User"
    valid_email = "test@mozilla.com"

    assert not helpers.is_valid_email(invalid_email)
    assert helpers.is_valid_email(valid_email)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import hashlib
import uuid

import pytest

from contextlib import contextmanager
from unittest import mock

from mozphab import user

from mozphab.conduit import ConduitAPIError
from mozphab.exceptions import Error


@mock.patch("mozphab.user.USER_INFO_FILE")
@mock.patch("mozphab.user.json")
def test_save_user_info(m_json, m_file, user_data):
    @contextmanager
    def with_open(*_, **__):
        yield None

    m_file.open = with_open
    # create file
    m_file.exists.return_value = False
    installation_id = str(uuid.uuid4())
    user_info = dict(
        is_employee=True,
        user_code=str(uuid.uuid4()),
        installation_id=installation_id,
        last_check=1,
    )
    user_data.save_user_info(**user_info)
    m_json.dump.assert_called_once_with(user_info, None, sort_keys=True, indent=2)

    # update file
    m_file.exists.return_value = True
    m_json.reset_mock()
    m_json.load.return_value = user_info
    new_user_code = str(uuid.uuid4())
    user_data.save_user_info(user_code=new_user_code)
    m_json.dump.assert_called_once_with(
        dict(
            is_employee=True,
            user_code=new_user_code,
            installation_id=installation_id,
            last_check=1,
        ),
        None,
        sort_keys=True,
        indent=2,
    )


def test_is_data_collected(user_data):
    user_data.update_from_dict(
        dict(
            is_employee=None,
            user_code="ff455518-16ae-4a89-b600-f08e30c25ad2",
            installation_id="5e3c5706-09c2-41a4-8216-e5e463b032b9",
            last_check=1,
        )
    )
    assert not user_data.is_data_collected
    user_data.update_from_dict(
        dict(
            is_employee=True,
            user_code=None,
            installation_id="5e3c5706-09c2-41a4-8216-e5e463b032b9",
            last_check=1,
        )
    )
    assert not user_data.is_data_collected
    user_data.update_from_dict(
        dict(
            is_employee=False,
            user_code="ff455518-16ae-4a89-b600-f08e30c25ad2",
            installation_id=None,
            last_check=1,
        )
    )
    assert not user_data.is_data_collected
    user_data.update_from_dict(
        dict(
            is_employee=True,
            user_code="ff455518-16ae-4a89-b600-f08e30c25ad2",
            installation_id="5e3c5706-09c2-41a4-8216-e5e463b032b9",
            last_check=None,
        )
    )
    assert not user_data.is_data_collected
    user_data.update_from_dict(
        dict(
            is_employee=False,
            user_code="ff455518-16ae-4a89-b600-f08e30c25ad2",
            installation_id="5e3c5706-09c2-41a4-8216-e5e463b032b9",
            last_check=1,
        )
    )
    assert user_data.is_data_collected


@mock.patch("mozphab.user.hashlib")
@mock.patch("mozphab.user.UserData.whoami")
@mock.patch("mozphab.user.UserData.save_user_info")
@mock.patch("mozphab.user.USER_INFO_FILE")
@mock.patch("mozphab.user.time")
def test_set_user_data(m_time, m_file, m_save, m_whoami, m_hashlib, user_data):
    m_save.return_value = False
    m_file.exists.return_value = True
    m_time.time.return_value = user.EMPLOYEE_CHECK_FREQUENCY - 1
    # all data saved in file, no need to update
    user_data.update_from_dict(
        dict(is_employee=False, user_code="u", installation_id="i", last_check=2)
    )
    assert not user_data.set_user_data()
    m_whoami.assert_not_called()

    # return None if `from_file_only` == True and one of the info is not not saved
    user_data.is_employee = None
    assert not user_data.set_user_data(from_file_only=True)
    user_data.is_employee = True
    user_data.user_code = None
    assert not user_data.set_user_data(from_file_only=True)
    user_data.user_code = "user_code00000000000000000000000"
    user_data.installation_id = None
    assert not user_data.set_user_data(from_file_only=True)
    user_data.installation_id = "installation11111111111111111111"

    # return None if whoami() failed
    user_data.is_employee = None
    m_whoami.return_value = None
    assert not user_data.set_user_data()

    user_data.is_employee = None
    m_whoami.return_value = {"email": None}
    assert (
        not user_data.set_user_data()
    ), "Status should not be updated without a valid email."

    # Update user_data, not employee
    m_whoami.side_effect = (dict(email="someemail", is_employee=False),)
    hexdigest = mock.Mock()
    hashed_email = hashlib.md5("someemail".encode("utf-8")).hexdigest()
    user_code = str(uuid.UUID(hashed_email))
    hexdigest.hexdigest.return_value = hashed_email
    m_hashlib.md5.return_value = hexdigest
    m_time.time.return_value = 123
    assert user_data.set_user_data()
    assert (
        dict(
            user_code=user_code,
            is_employee=False,
            installation_id="installation11111111111111111111",
            last_check=123,
        )
        == user_data.to_dict()
    )

    # Create user_data file, employee
    m_file.exists.return_value = False
    m_whoami.side_effect = (dict(email="someemail", is_employee=True),)
    m_save.return_value = True
    user_data.installation_id = None
    assert user_data.set_user_data()
    user_data_dict = user_data.to_dict()
    assert (
        user_data_dict["is_employee"] is True
    ), "`whoami` showing user is employee should be reflected in user data."

    assert isinstance(
        user_data_dict["user_code"], str
    ), "`user_code` should be a `str`."
    assert (
        len(user_data_dict["user_code"]) == 36
    ), "`user_code` should be a 36-character `str`."
    assert "-" in user_data_dict["user_code"], "`user_code` should include hyphens."

    assert isinstance(
        user_data_dict["installation_id"], str
    ), "`installation_id` should be a `str`."
    assert (
        len(user_data_dict["installation_id"]) == 36
    ), "`installation_id` should be a 36-character `str`."
    assert (
        "-" in user_data_dict["installation_id"]
    ), "`installation_id` should include hyphens."

    # No file, read_from_file_only
    assert not user_data.set_user_data(from_file_only=True)


@mock.patch("mozphab.user.bmo")
@mock.patch("mozphab.user.conduit")
def test_whoami(m_conduit, m_bmo, user_data):
    # return None if conduit.whoami() raises
    m_conduit.whoami.side_effect = ConduitAPIError
    assert user_data.whoami() is None

    # An employee based on @mozilla.com email
    m_conduit.whoami.side_effect = None
    m_conduit.whoami.return_value = dict(primaryEmail="someemail@mozilla.com")
    assert user_data.whoami() == dict(email="someemail@mozilla.com", is_employee=True)

    # An employee based on @getpocket.com email
    m_conduit.whoami.side_effect = None
    m_conduit.whoami.return_value = dict(primaryEmail="someemail@getpocket.com")
    assert user_data.whoami() == dict(email="someemail@getpocket.com", is_employee=True)

    # An employee based on @mozillafoundation.com email
    m_conduit.whoami.side_effect = None
    m_conduit.whoami.return_value = dict(primaryEmail="someemail@mozillafoundation.org")
    assert user_data.whoami() == dict(
        email="someemail@mozillafoundation.org", is_employee=True
    )

    # Not employee as BMO.whoami failed
    m_conduit.whoami.return_value = dict(primaryEmail="some@email.com")
    m_bmo.whoami.return_value = None
    assert user_data.whoami() == dict(email="some@email.com", is_employee=False)

    # Not employee as not in mozilla-employee-confidential group
    m_conduit.whoami.return_value = dict(primaryEmail="some@email.com")
    m_bmo.whoami.return_value = dict(name="nvm@email.com", groups=["some-group"])
    assert user_data.whoami() == dict(email="some@email.com", is_employee=False)

    # An employee as in mozilla-employee-confidential group
    m_bmo.whoami.return_value = dict(
        name="someemail", groups=["mozilla-employee-confidential"]
    )
    assert user_data.whoami() == dict(email="some@email.com", is_employee=True)

    m_conduit.whoami.return_value = dict(primaryEmail=None)
    assert user_data.whoami() == dict(
        email=None, is_employee=False
    ), "When `primaryEmail` is empty, `is_employee` is False and doesn't fail."

    # whoami raises an error if BMO.whoami raises Error
    m_conduit.whoami.return_value = dict(primaryEmail="some@email.com")
    m_bmo.whoami.side_effect = Error
    with pytest.raises(Error):
        user_data.whoami()


def test_is_bad_uuid():
    assert not user.is_bad_uuid(
        "randomkey", None
    ), "Irrelevant key should return False."
    assert not user.is_bad_uuid(
        "randomkey", "asdf"
    ), "Irrelevant key should return False."

    assert not user.is_bad_uuid(
        "installation_id", None
    ), "Empty value is not a bad UUID."
    assert not user.is_bad_uuid("user_code", None), "Empty value is not a bad UUID."

    assert not user.is_bad_uuid(
        "installation_id", "blah"
    ), "Bad UUID can only be 32-chars."
    assert not user.is_bad_uuid("user_code", "blah"), "Bad UUID can only be 32-chars."

    assert user.is_bad_uuid(
        "user_code", "a" * 32
    ), "32-char `str` with appropriate key is bad UUID."
    assert user.is_bad_uuid(
        "installation_id", "a" * 32
    ), "32-char `str` with appropriate key is bad UUID."


def test_format_uuid():
    assert (
        user.format_uuid("83a4fbef668c4b168dc7f62427c1b855")
        == "83a4fbef-668c-4b16-8dc7-f62427c1b855"
    ), "Bad UUID should be converted to a valid one."


def test_bad_uuid_rewritten(user_data):
    user_data.update_from_dict(
        {
            "installation_id": "83a4fbef668c4b168dc7f62427c1b855",
            "user_code": "83a4fbef668c4b168dc7f62427c1b855",
        }
    )
    assert (
        user_data.installation_id == "83a4fbef-668c-4b16-8dc7-f62427c1b855"
    ), "`installation_id` should be rewritten to good UUID on update."
    assert (
        user_data.user_code == "83a4fbef-668c-4b16-8dc7-f62427c1b855"
    ), "`user_code` should be rewritten to good UUID on update."


@mock.patch("mozphab.user.USER_INFO_FILE")
@mock.patch("mozphab.user.UserData.whoami")
def test_new_telemetry_ids_are_valid(m_whoami, m_file, user_data):
    m_whoami.side_effect = ({"email": "someemail", "is_employee": False},)
    m_file.exists.return_value = False
    assert (
        user_data.installation_id is None
    ), "Sanity check `installation_id` starts as `None`."

    user_data.set_user_data()

    assert (
        len(user_data.installation_id) == 36
    ), "`installation_id` should have len of 36."
    assert "-" in user_data.installation_id, "`installation_id` should contain hyphens."

    assert len(user_data.user_code) == 36, "`user_code` should have len of 36."
    assert "-" in user_data.user_code, "`user_code` should contain hyphens."



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import signal
import sys
import threading
import time

from contextlib import contextmanager
from mozphab import environment


def clear_terminal_line():
    if environment.HAS_ANSI:
        sys.stdout.write("\r\033[K")  # move to start of line, erase to end of line
        sys.stdout.flush()


def signal_sigint(self, *args):
    print("\nCancelled")
    raise KeyboardInterrupt()


signal.signal(signal.SIGINT, signal_sigint)


class Spinner(threading.Thread):
    def __init__(self, message):
        super().__init__()
        self.message = message
        self.daemon = True
        self.running = False

    def run(self):
        self.running = True

        if not environment.HAS_ANSI:
            sys.stdout.write("%s  " % self.message)

        spinner = ["-", "\\", "|", "/"]
        spin = 0
        try:
            while self.running:
                if environment.HAS_ANSI:
                    sys.stdout.write("%s %s\r" % (self.message, spinner[spin]))
                else:
                    sys.stdout.write(chr(8) + spinner[spin])
                sys.stdout.flush()
                spin = (spin + 1) % len(spinner)
                time.sleep(0.2)
        finally:
            if environment.HAS_ANSI:
                clear_terminal_line()
            else:
                sys.stdout.write(chr(8) + " \n")


@contextmanager
def wait_message(message):
    if not environment.SHOW_SPINNER:
        yield
        return

    spinner = Spinner(message)
    spinner.start()
    try:
        yield
    finally:
        spinner.running = False
        spinner.join()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pytest
import uuid

from unittest import mock

from mozphab import telemetry
from mozphab.bmo import BMOAPIError


@pytest.fixture
def get_telemetry():
    return telemetry.Telemetry()


class Args:
    def __init__(self, needs_repo=False, command="submit", force=None, force_vcs=None):
        self.needs_repo = needs_repo
        self.command = command
        self.no_arc = True
        if force_vcs is not None:
            self.force_vcs = force_vcs
        if force is not None:
            self.force = force


def test_vcs(get_telemetry):
    class Repo:
        def __init__(self, vcs, vcs_version):
            self.vcs = vcs
            self.vcs_version = vcs_version

    t = get_telemetry
    t.set_vcs(Repo("hg", "5.2"))
    assert "hg" == t.environment.vcs.name.test_get_value()
    assert "5.2" == t.environment.vcs.version.test_get_value()


@mock.patch("mozphab.telemetry.platform")
@mock.patch("mozphab.telemetry.distro")
def test_set_os(m_distro, m_platform, get_telemetry):
    t = get_telemetry
    m_platform.uname.side_effect = (("Linux", "node", "release", None, None, None),)
    m_distro.linux_distribution.side_effect = (("debian", "2020.1", None),)
    t._set_os()
    assert "debian 2020.1" == t.environment.distribution_version.test_get_value()

    m_platform.uname.side_effect = (("Windows", "node", "release", None, None, None),)
    m_platform.win32_ver.side_effect = (("10", "10.0.18362", "", "multiprocessor"),)
    t._set_os()
    assert "10.0.18362" == t.environment.distribution_version.test_get_value()

    m_platform.uname.side_effect = (("Darwin", "node", "release", None, None, None),)
    m_platform.mac_ver.side_effect = (("10.15.3", ("", "", ""), "x86_64"),)
    t._set_os()
    assert "10.15.3" == t.environment.distribution_version.test_get_value()

    m_platform.uname.side_effect = (("Something", "node", "release", None, None, None),)
    t._set_os()
    assert "release" == t.environment.distribution_version.test_get_value()


@mock.patch("mozphab.telemetry.platform")
def test_set_python(m_platform, get_telemetry):
    m_platform.python_version.side_effect = ("3.7.6",)
    t = get_telemetry
    t._set_python()
    assert "3.7.6" == t.environment.python_version.test_get_value()


@mock.patch("mozphab.telemetry.config")
@mock.patch("mozphab.telemetry.user_data")
@mock.patch("mozphab.telemetry.prompt")
def test_update_user_data(m_prompt, m_user_data, m_config):
    m_config.configure_mock(telemetry_enabled=False)

    # user_data not updated
    m_user_data.set_user_data.return_value = False
    telemetry.update_user_data()
    assert not m_config.telemetry_enabled
    m_config.write.assert_not_called()

    # not instantiated, BMOAPIError
    m_user_data.set_user_data.side_effect = BMOAPIError
    with pytest.raises(BMOAPIError):
        telemetry.update_user_data()

    # switched off, not instantiated, user data retrieved from BMO, employee
    m_user_data.set_user_data.side_effect = None
    m_user_data.set_user_data.return_value = True
    m_user_data.configure_mock(
        is_employee=True,
        is_data_collected=True,
    )
    m_config.configure_mock(telemetry_enabled=True)
    telemetry.update_user_data()
    assert m_config.telemetry_enabled
    m_config.write.assert_called_once()

    # switched off, not instantiated, user data retrieved from BMO, not employee
    m_user_data.is_employee = False
    m_config.telemetry_enabled = False

    # not opt-in
    m_config.write.reset_mock()
    m_prompt.return_value = "No"
    telemetry.update_user_data()
    assert not m_config.telemetry_enabled
    m_config.write.assert_called_once()

    # opt-in
    m_config.write.reset_mock()
    m_prompt.return_value = "Yes"
    telemetry.update_user_data()
    assert m_config.telemetry_enabled
    m_config.write.assert_called_once()


@mock.patch("mozphab.telemetry.update_user_data")
@mock.patch("mozphab.telemetry.user_data")
@mock.patch("mozphab.telemetry.config")
def test_set_metrics(m_config, m_user_data, m_update_user_data, get_telemetry):
    t = get_telemetry
    t._set_os = mock.Mock()
    t._set_python = mock.Mock()
    m_config.configure_mock(telemetry_enabled=True)

    # switched on in `update_user_data`
    m_config.configure_mock(telemetry_enabled=False)
    m_update_user_data.side_effect = None
    m_update_user_data.return_value = True
    m_user_data.configure_mock(
        is_data_collected=True,
        installation_id="0000aabb-bbaa-0000-aabb-0000bbaa0000",
        user_code="1111ffee-eeff-1111-ffee-1111eeff1111",
    )
    # telemetry is enabled in `update_user_data`
    t.set_metrics(Args(needs_repo=True))
    t._set_os.assert_called_once()
    t._set_python.assert_called_once()
    assert "submit" == t.usage.command.test_get_value()
    assert t.usage.override_switch.test_get_value() is False
    # `command_time.start()` has been called, but not stop, it has no value yet
    assert t.usage.command_time.test_get_value() is None
    assert (
        uuid.UUID("0000aabb-bbaa-0000-aabb-0000bbaa0000")
        == t.user.installation.test_get_value()
    )
    assert (
        uuid.UUID("1111ffee-eeff-1111-ffee-1111eeff1111") == t.user.id.test_get_value()
    )


@mock.patch("mozphab.telemetry.Telemetry")
@mock.patch("mozphab.telemetry.update_user_data")
@mock.patch("mozphab.telemetry.user_data")
@mock.patch("mozphab.telemetry.config")
def test_configure_telemetry(
    m_config,
    m_user_data,
    m_update_user_data,
    m_telemetry,
):
    # install-certificate -> disabled
    telemetry.configure_telemetry(Args(command="install-certificate"))
    assert isinstance(telemetry.telemetry(), telemetry.TelemetryDisabled)

    # self-update -> disabled
    telemetry.configure_telemetry(Args(command="self-update"))
    assert isinstance(telemetry.telemetry(), telemetry.TelemetryDisabled)

    # off -> disabled
    m_config.configure_mock(telemetry_enabled=False)
    m_user_data.configure_mock(is_data_collected=True)
    telemetry.configure_telemetry(Args())
    assert isinstance(telemetry.telemetry(), telemetry.TelemetryDisabled)

    # on, no user data -> disabled
    m_config.configure_mock(telemetry_enabled=True)
    m_user_data.configure_mock(is_data_collected=False)
    telemetry.configure_telemetry(Args())
    assert isinstance(telemetry.telemetry(), telemetry.TelemetryDisabled)

    # on, bmo error -> disabled
    m_config.configure_mock(telemetry_enabled=True)
    m_update_user_data.side_effect = BMOAPIError
    telemetry.configure_telemetry(Args(needs_repo=True))
    assert isinstance(telemetry.telemetry(), telemetry.TelemetryDisabled)

    # on --> enabled
    m_config.configure_mock(telemetry_enabled=True)
    m_user_data.configure_mock(is_data_collected=True)
    telemetry.configure_telemetry(Args())
    m_telemetry.assert_called_once()
    assert not isinstance(telemetry.telemetry(), telemetry.TelemetryDisabled)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# coding=utf-8

from unittest import mock
import textwrap

from .conftest import assert_attributes

from mozphab import environment
from mozphab.diff import Diff


class Args:
    def __init__(self, less_context=False):
        self.lesscontext = less_context


@mock.patch("mozphab.git.Git._file_size")
@mock.patch("mozphab.git.Git._cat_file")
@mock.patch("mozphab.git.Git.git_out")
def test_create(m_git_out, m_cat_file, m_file_size, git):
    raw = (
        "000000 100644 0000000000000000000000000000000000000000 "
        "78981922613b2afb6025042ff6bd878ac1994e85 A\x00a"
    )
    diff = Diff()
    m_cat_file.side_effect = (b"a\n",)
    m_file_size.return_value = 5
    git.args = Args()

    change = git._parse_diff_change(raw, diff)
    m_git_out.assert_not_called()
    assert change.file_type.name == "TEXT"
    assert change.kind.name == "ADD"
    assert change.cur_mode == "100644"
    assert len(change.hunks) == 1
    assert_attributes(
        change.hunks[0],
        dict(
            old_off=0,
            old_len=0,
            new_off=1,
            new_len=1,
            old_eof_newline=True,
            new_eof_newline=True,
            added=1,
            deleted=0,
            corpus="+a\n",
        ),
    )


@mock.patch("mozphab.git.Git._file_size")
@mock.patch("mozphab.git.Git._cat_file")
@mock.patch("mozphab.git.Git.git_out")
def test_change_file(m_git_out, m_cat_file, m_file_size, git):
    raw = (
        "100644 100644 78981922613b2afb6025042ff6bd878ac1994e85 "
        "422c2b7ab3b3c668038da977e4e93a5fc623169c M\x00a"
    )
    diff = Diff()
    m_cat_file.side_effect = (b"a\n", b"a\nb\n")
    m_git_out.return_value = b"""\
diff --git a/78981922613b2afb6025042ff6bd878ac1994e85 \
b/422c2b7ab3b3c668038da977e4e93a5fc623169c
index 7898192..422c2b7 100644
--- a/78981922613b2afb6025042ff6bd878ac1994e85\n+++ \
b/422c2b7ab3b3c668038da977e4e93a5fc623169c
@@ -1 +1,2 @@
 a
+b"""
    m_file_size.return_value = 5
    git.args = Args()

    change = git._parse_diff_change(raw, diff)
    m_git_out.assert_called_once_with(
        [
            "diff",
            "--submodule=short",
            "--no-ext-diff",
            "--no-color",
            "--no-textconv",
            "-U%s" % environment.MAX_CONTEXT_SIZE,
            "78981922613b2afb6025042ff6bd878ac1994e85",
            "422c2b7ab3b3c668038da977e4e93a5fc623169c",
        ],
        expect_binary=True,
    )
    assert change.file_type.name == "TEXT"
    assert change.kind.name == "CHANGE"
    assert change.old_mode is None
    assert change.cur_mode is None
    assert change.old_path == "a"
    assert len(change.hunks) == 1
    assert_attributes(
        change.hunks[0],
        dict(
            old_off=1,
            old_len=1,
            new_off=1,
            new_len=2,
            old_eof_newline=True,
            new_eof_newline=True,
            added=1,
            deleted=0,
            corpus=" a\n+b",
        ),
    )


@mock.patch("mozphab.git.Git._file_size")
@mock.patch("mozphab.git.Git._cat_file")
@mock.patch("mozphab.git.Git.git_out")
def test_create_empty(m_git_out, m_cat_file, m_file_size, git):
    raw = (
        "000000 100644 0000000000000000000000000000000000000000 "
        "78981922613b2afb6025042ff6bd878ac1994e85 A\x00a"
    )
    diff = Diff()
    m_cat_file.side_effect = (b"",)
    m_file_size.return_value = 0
    git.args = Args()

    change = git._parse_diff_change(raw, diff)
    m_git_out.assert_not_called()
    assert change.file_type.name == "TEXT"
    assert change.hunks == []
    assert change.kind.name == "ADD"
    assert change.cur_mode == "100644"


@mock.patch("mozphab.git.Git._file_size")
@mock.patch("mozphab.git.Git._cat_file")
@mock.patch("mozphab.git.Git.git_out")
def test_change_empty(m_git_out, m_cat_file, m_file_size, git):
    raw = (
        "100644 100755 0000000000000000000000000000000000000000 "
        "78981922613b2afb6025042ff6bd878ac1994e85 M\x00a"
    )
    diff = Diff()
    m_cat_file.side_effect = (b"",)
    m_file_size.return_value = 0
    git.args = Args()

    change = git._parse_diff_change(raw, diff)
    m_git_out.assert_not_called()
    assert (
        change.file_type.name == "TEXT"
    ), "The file should be recognized as a TEXT file"
    assert (
        change.hunks == []
    ), "`change.hunks` should be an empty list when parsing empty files"
    assert (
        change.kind.name == "CHANGE"
    ), "File mode changes should be recognized as a CHANGE"
    assert (
        change.old_mode == "100644"
    ), "The file's original mode should be set to 100644"
    assert (
        change.cur_mode == "100755"
    ), "The file's current mode should be set to 100755"


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
@mock.patch("mozphab.mercurial.Mercurial._get_file_meta")
@mock.patch("mozphab.mercurial.Mercurial._get_parent")
@mock.patch("mozphab.mercurial.Mercurial._get_file_modes")
@mock.patch("uuid.uuid4")
def test_change_empty_hg(
    m_uuid4, m_get_file_modes, m_get_parent, m_get_file_meta, m_hg_out, hg
):
    commit = {
        "name": "78981922613b",
        "node": "78981922613b2afb6025042ff6bd878ac1994e85",
        "orig-node": "78981922613b2afb6025042ff6bd878ac1994e85",
        "parent": "422c2b7ab3b3",
        "title": "test",
        "body": "test",
    }
    m_get_parent.return_value = "422c2b7ab3b3c668038da977e4e93a5fc623169c"
    m_get_file_modes.return_value = {
        "fn": {
            "old_mode": "100644",
            "new_mode": "100755",
        },
    }
    m_uuid4.side_effect = [
        mock.Mock(hex="abc123"),
        mock.Mock(hex="def456"),
    ]
    m_hg_out.side_effect = [
        "--abc123----def456----abc123----def456--fn--abc123----def456----abc123--",
    ]
    m_get_file_meta.side_effect = [
        dict(
            binary=False,
            bin_body=b"",
            body="",
            file_size=0,
        ),
        dict(
            binary=False,
            bin_body=b"",
            body="",
            file_size=0,
        ),
    ]

    hg.args = Args()
    diff = hg.get_diff(commit)
    change = diff.changes.get("fn")

    assert (
        change.kind.name == "CHANGE"
    ), "File mode changes should be recognized as a CHANGE"
    assert (
        change.old_mode == "100644"
    ), "The file's original mode should be set to 100644"
    assert (
        change.cur_mode == "100755"
    ), "The file's current mode should be set to 100755"
    assert (
        change.hunks == []
    ), "`change.hunks` should be an empty list when parsing empty files"


@mock.patch("mozphab.git.Git._file_size")
@mock.patch("mozphab.git.Git._cat_file")
@mock.patch("mozphab.git.Git.git_out")
def test_delete_file(m_git_out, m_cat_file, m_file_size, git):
    raw = (
        "100644 000000 61780798228d17af2d34fce4cfbdf35556832472 "
        "0000000000000000000000000000000000000000 D\x00a"
    )
    diff = Diff()
    m_cat_file.side_effect = (b"a\nb\n",)
    m_file_size.return_value = 5
    git.args = Args()

    change = git._parse_diff_change(raw, diff)
    m_git_out.assert_not_called()
    assert change.file_type.name == "TEXT"
    assert_attributes(
        change.hunks[0],
        dict(
            old_off=1,
            old_len=2,
            new_off=0,
            new_len=0,
            old_eof_newline=True,
            new_eof_newline=True,
            added=0,
            deleted=2,
            corpus="-a\n-b\n",
        ),
    )


@mock.patch("mozphab.git.Git._file_size")
@mock.patch("mozphab.git.Git._cat_file")
@mock.patch("mozphab.git.Git.git_out")
def test_delete_empty_file(m_git_out, m_cat_file, m_file_size, git):
    raw = (
        "100644 000000 61780798228d17af2d34fce4cfbdf35556832472 "
        "0000000000000000000000000000000000000000 D\x00a"
    )
    diff = Diff()
    m_cat_file.side_effect = (b"",)
    m_file_size.return_value = 0
    git.args = Args()

    change = git._parse_diff_change(raw, diff)
    m_git_out.assert_not_called()
    assert change.file_type.name == "TEXT"
    assert change.hunks == []


@mock.patch("mozphab.git.Git._file_size")
@mock.patch("mozphab.git.Git._cat_file")
@mock.patch("mozphab.git.Git.git_out")
def test_recognize_binary(m_git_out, m_cat_file, m_file_size, git):
    raw = (
        "000000 100644 0000000000000000000000000000000000000000 "
        "21be03052ed0c8dc31dff33eeb9275430241a727 A\x00sample.bin"
    )
    diff = Diff()
    content = b"\x08\x00\x00\x10"
    m_cat_file.side_effect = (content,)
    m_file_size.return_value = 5
    git.args = Args()

    change = git._parse_diff_change(raw, diff)
    m_git_out.assert_not_called()
    assert change.file_type.name == "BINARY"
    assert change.uploads == [
        dict(type="old", value=b"", mime="application/octet-stream", phid=None),
        dict(type="new", value=content, mime="application/octet-stream", phid=None),
    ]
    assert not change.hunks


@mock.patch("mozphab.git.Git._file_size")
@mock.patch("mozphab.git.Git._cat_file")
@mock.patch("mozphab.git.Git.git_out")
def test_recognize_long_text_as_binary(m_git_out, m_cat_file, m_file_size, git):
    raw = (
        "000000 100644 0000000000000000000000000000000000000000 "
        "78981922613b2afb6025042ff6bd878ac1994e85 A\x00a"
    )
    diff = Diff()
    content = b"a\n"
    m_cat_file.side_effect = (content,)
    m_file_size.return_value = environment.MAX_TEXT_SIZE + 1
    git.args = Args()

    change = git._parse_diff_change(raw, diff)
    m_git_out.assert_not_called()
    assert change.file_type.name == "BINARY"
    assert change.uploads == [
        dict(type="old", value=b"", mime="", phid=None),
        dict(type="new", value=content, mime="", phid=None),
    ]
    assert not change.hunks


@mock.patch("mozphab.git.Git._file_size")
@mock.patch("mozphab.git.Git._cat_file")
@mock.patch("mozphab.git.Git.git_out")
def test_less_context(m_git_out, m_cat_file, m_file_size, git):
    raw = (
        "100644 100644 78981922613b2afb6025042ff6bd878ac1994e85 "
        "422c2b7ab3b3c668038da977e4e93a5fc623169c M\x00a"
    )
    diff = Diff()
    m_cat_file.side_effect = (b"a\n", b"a\nb\n")
    m_git_out.return_value = b"""\
diff --git a/78981922613b2afb6025042ff6bd878ac1994e85 \
b/422c2b7ab3b3c668038da977e4e93a5fc623169c
index 7898192..422c2b7 100644
--- a/78981922613b2afb6025042ff6bd878ac1994e85\n+++ \
b/422c2b7ab3b3c668038da977e4e93a5fc623169c
@@ -1 +1,2 @@
 a
+b"""
    m_file_size.return_value = 5
    git.args = Args(less_context=True)

    git._parse_diff_change(raw, diff)
    m_git_out.assert_called_once_with(
        [
            "diff",
            "--submodule=short",
            "--no-ext-diff",
            "--no-color",
            "--no-textconv",
            "-U100",
            "78981922613b2afb6025042ff6bd878ac1994e85",
            "422c2b7ab3b3c668038da977e4e93a5fc623169c",
        ],
        expect_binary=True,
    )

    git.args = Args(less_context=False)
    m_file_size.return_value = environment.MAX_CONTEXT_SIZE + 1
    m_cat_file.side_effect = (b"a\n", b"a\nb\n")
    m_git_out.reset_mock()

    git._parse_diff_change(raw, diff)
    m_git_out.assert_called_once_with(
        [
            "diff",
            "--submodule=short",
            "--no-ext-diff",
            "--no-color",
            "--no-textconv",
            "-U100",
            "78981922613b2afb6025042ff6bd878ac1994e85",
            "422c2b7ab3b3c668038da977e4e93a5fc623169c",
        ],
        expect_binary=True,
    )


def test_multiple_hunks():
    git_diff = textwrap.dedent(
        """
        diff --git a/fn b/fn
        --- a/fn
        +++ b/fn
        @@ -4,3 +4,2 @@ c
        d
        -e
        f
        @@ -11,3 +10,2 @@ j
        k
        -l
        m
        @@ -25,2 +21,1 @@ x
        y
        -z
        """
    )

    change = Diff.Change("x")
    change.from_git_diff(git_diff)

    assert len(change.hunks) == 3
    assert_attributes(
        change.hunks[0],
        dict(
            old_off=4,
            old_len=3,
            new_off=4,
            new_len=2,
            old_eof_newline=True,
            new_eof_newline=True,
            added=0,
            deleted=1,
            corpus="d\n-e\nf\n",
        ),
    )
    assert_attributes(
        change.hunks[1],
        dict(
            old_off=11,
            old_len=3,
            new_off=10,
            new_len=2,
            old_eof_newline=True,
            new_eof_newline=True,
            added=0,
            deleted=1,
            corpus="k\n-l\nm\n",
        ),
    )
    assert_attributes(
        change.hunks[2],
        dict(
            old_off=25,
            old_len=2,
            new_off=21,
            new_len=1,
            old_eof_newline=True,
            new_eof_newline=True,
            added=0,
            deleted=1,
            corpus="y\n-z\n",
        ),
    )


def test_set_as_binary():
    change = Diff.Change("x")
    change.set_as_binary(
        a_body=b"a",
        a_mime="pdf/",
        b_body=b"b",
        b_mime="pdf/",
    )
    assert change.binary
    assert change.uploads == [
        {"type": "old", "value": b"a", "mime": "pdf/", "phid": None},
        {"type": "new", "value": b"b", "mime": "pdf/", "phid": None},
    ]
    assert change.file_type.name == "BINARY"

    change = Diff.Change("x")
    change.set_as_binary(
        a_body=b"a",
        a_mime="image/jpeg",
        b_body=b"b",
        b_mime="pdf/",
    )
    assert change.file_type.name == "IMAGE"

    change = Diff.Change("x")
    change.set_as_binary(
        a_body=b"a",
        a_mime="image/jpeg",
        b_body=b"b",
        b_mime="pdf/",
    )
    assert change.file_type.name == "IMAGE"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozphab import environment

from mozphab.conduit import conduit
from mozphab.config import config
from mozphab.exceptions import Error
from mozphab.helpers import (
    augment_commits_from_body,
    BLOCKING_REVIEWERS_RE,
    parse_arc_diff_rev,
    prompt,
    move_drev_to_original,
    revision_title_from_commit,
    strip_differential_revision,
    update_commit_title_previews,
)
from mozphab.logger import logger
from mozphab.spinner import wait_message
from mozphab.telemetry import telemetry

# The DEFAULT_UPDATE_MESSAGE is only required when using arc to submit.
DEFAULT_UPDATE_MESSAGE = "Revision updated."

PHABRICATOR_URLS = {
    "https://phabricator.services.mozilla.com/": "Phabricator",
    "https://phabricator-dev.allizom.org/": "Phabricator-Dev",
}

ARC_COMMIT_DESC_TEMPLATE = """
{title}

Summary:
{body}

{depends_on}

Test Plan:

Reviewers: {reviewers}

Subscribers:

Bug #: {bug_id}
""".strip()


def morph_blocking_reviewers(commits):
    """Automatically fix common typo by replacing r!user with r=user!"""

    def morph_reviewer(matchobj):
        if matchobj.group(1) == "r!":
            nick = matchobj.group(2)

            # strip trailing , or . so we can put it back later
            if nick[-1] in (",", "."):
                suffix = nick[-1]
                nick = nick[:-1]
            else:
                suffix = ""

            # split on comma to support r!a,b -> r=a!,b!
            return "r=%s%s" % (
                ",".join(["%s!" % n.rstrip("!") for n in nick.split(",")]),
                suffix,
            )
        else:
            return matchobj.group(0)

    for commit in commits:
        commit["title"] = BLOCKING_REVIEWERS_RE.sub(morph_reviewer, commit["title"])


def arc_message(template_vars):
    """Build arc commit desc message from the template"""

    # Map `None` to an empty string.
    for name in list(template_vars.keys()):
        if template_vars[name] is None:
            template_vars[name] = ""

    # `depends_on` is optional.
    if "depends_on" not in template_vars:
        template_vars["depends_on"] = ""

    message = ARC_COMMIT_DESC_TEMPLATE.format(**template_vars)
    logger.debug("--- arc message\n%s\n---" % message)
    return message


def amend_revision_url(body, new_url):
    """Append or replace the Differential Revision URL in a commit body."""
    body = strip_differential_revision(body)
    if body:
        body += "\n"
    body += "\nDifferential Revision: %s" % new_url
    return body


def show_commit_stack(
    commits,
    validate=True,
    show_rev_urls=False,
    show_updated_only=False,
):
    """Log the commit stack in a human readable form."""

    # keep output in columns by sizing the action column to the longest rev + 1 ("D")
    max_len = (
        max(len(c.get("rev-id", "") or "") for c in commits) + 1 if commits else ""
    )
    action_template = "(%" + str(max_len) + "s)"

    if validate:
        # preload all revisions
        ids = [int(c["rev-id"]) for c in commits if c.get("rev-id")]
        if ids:
            with wait_message("Loading existing revisions..."):
                revisions = conduit.get_revisions(ids=ids)

            # preload diffs
            with wait_message("Loading diffs..."):
                diffs = conduit.get_diffs(
                    phids=[r["fields"]["diffPHID"] for r in revisions]
                )

    for commit in reversed(commits):
        if show_updated_only and not commit["submit"]:
            continue

        revision_is_closed = False
        revision_is_wip = False
        bug_id_changed = False
        is_author = True
        revision = None

        if commit.get("rev-id"):
            action = action_template % ("D" + commit["rev-id"])

            if validate:
                revisions = conduit.get_revisions(ids=[int(commit["rev-id"])])
                if revisions:
                    revision = revisions[0]
                    fields = revision["fields"]

                    # WIP if either in changes-planned state, or if the revision has the
                    # `draft` flag set.
                    revision_is_wip = (
                        fields["status"]["value"] == "changes-planned"
                        or fields["isDraft"]
                    )

                    # Check if target bug ID is the same as in the Phabricator revision
                    bug_id_changed = fields.get("bugzilla.bug-id") and (
                        commit["bug-id"] != fields["bugzilla.bug-id"]
                    )

                    # Check if revision is closed
                    revision_is_closed = fields["status"]["closed"]

                    # Check if comandeering is required
                    with wait_message("Figuring out who you are..."):
                        whoami = conduit.whoami()
                    if "authorPHID" in fields and (
                        fields["authorPHID"] != whoami["phid"]
                    ):
                        is_author = False

                    # Any reviewers added to a revision without them?
                    reviewers_added = bool(
                        not revision["attachments"]["reviewers"]["reviewers"]
                        and commit["reviewers"]["granted"]
                    )

                    # if SHA1 hasn't changed
                    # and we're not changing the WIP or draft status
                    # and we're not adding reviewers to a revision without reviewers
                    # and we're not changing the bug-id
                    diff_phid = fields["diffPHID"]
                    diff_commits = diffs[diff_phid]["attachments"]["commits"]["commits"]
                    sha1_changed = (
                        not diff_commits
                        or commit["node"] != diff_commits[0]["identifier"]
                    )
                    if (
                        not sha1_changed
                        and commit["wip"] == revision_is_wip
                        and not reviewers_added
                        and not bug_id_changed
                        and not revision_is_closed
                    ):
                        commit["submit"] = False

        else:
            action = action_template % "New"

        logger.info("%s %s %s", action, commit["name"], commit["title-preview"])
        if validate:
            if not commit["submit"]:
                logger.info(
                    " * This revision has not changed and will not be submitted."
                )
                continue

            if revision:
                if not commit["wip"] and revision_is_wip:
                    logger.warning(
                        '!! "Changes Planned" status will change to "Request Review"'
                    )
                if commit["wip"] and not revision_is_wip:
                    logger.warning(
                        '!! "Request Review" status will change to "Changes Planned"'
                    )

            if bug_id_changed:
                logger.warning(
                    "!! Bug ID in Phabricator revision will change from %s to %s",
                    revision["fields"]["bugzilla.bug-id"],
                    commit["bug-id"],
                )

            if not is_author:
                logger.warning(
                    "!! You don't own this revision. Normally, you should only\n"
                    '   update revisions you own. You can "Commandeer" this\n'
                    "   revision from the web interface if you want to become\n"
                    "   the owner."
                )

            if revision_is_closed:
                logger.warning(
                    "!! This revision is closed!\n"
                    "   It will be reopened if submission proceeds.\n"
                    "   You can stop now and refine the stack range."
                )

            if not commit["bug-id"]:
                logger.warning("!! Missing Bug ID")

            if commit["bug-id-orig"] and commit["bug-id"] != commit["bug-id-orig"]:
                logger.warning(
                    "!! Bug ID changed from %s to %s",
                    commit["bug-id-orig"],
                    commit["bug-id"],
                )

            if not commit["has-reviewers"]:
                logger.warning("!! Missing reviewers")
                if commit["wip"]:
                    logger.warning(
                        '   It will be submitted as "Changes Planned".\n'
                        "   Run submit again with --no-wip to prevent this."
                    )

        if show_rev_urls and commit["rev-id"]:
            logger.warning("-> %s/D%s", conduit.repo.phab_url, commit["rev-id"])


def make_blocking(reviewers):
    return ["%s!" % r.rstrip("!") for r in reviewers]


def remove_duplicates(reviewers):
    """Remove all duplicate items from the list.

    Args:
        reviewers: list of strings representing IRC nicks of reviewers.

    Returns: list of unique reviewers.

    Duplicates with excalamation mark are preferred.
    """
    unique = []
    nicks = []
    for reviewer in reviewers:
        nick = reviewer.lower().strip("!")
        if reviewer.endswith("!") and nick in nicks:
            nicks.remove(nick)
            unique = [r for r in unique if r.lower().strip("!") != nick]

        if nick not in nicks:
            nicks.append(nick)
            unique.append(reviewer)

    return unique


def update_commits_from_args(commits, args):
    """Modify commit description based on args and configuration.

    Args:
        commits: list of dicts representing the commits in commit stack
        args: argparse.ArgumentParser object. In this function we're using
            following attributes:
            * reviewer - list of strings representing reviewers
            * blocker - list of string representing blocking reviewers
            * bug - an integer representing bug id in Bugzilla

    Command args always overwrite commit desc.
    Overwriting reviewers rules
    (The "-r" are loaded into args.reviewer and "-R" into args.blocker):
        a) Recognize `r=` and `r?` from commit message title.
        b) Modify reviewers only if `-r` or `-R` is used in call arguments.
        c) Add new reviewers using `r=`.
        d) Do not modify reviewers already provided in the title (`r?` or `r=`).
        e) Delete reviewer if `-r` or `-R` is used and the reviewer is not mentioned.

    If no reviewers are provided by args and an `always_blocking` flag is set
    change all reviewers in title to blocking.
    """

    # Build list of reviewers from args.
    reviewers = list(set(args.reviewer)) if args.reviewer else []

    # Order might be changed here `list(set(["one", "two"])) == ['two', 'one']`
    reviewers.sort()
    blockers = [r.rstrip("!") for r in args.blocker] if args.blocker else []

    # User might use "-r <nick>!" to provide a blocking reviewer.
    # Add all such blocking reviewers to blockers list.
    blockers += [r.rstrip("!") for r in reviewers if r.endswith("!")]
    blockers = list(set(blockers))
    blockers.sort()

    # Remove blocking reviewers from reviewers list
    reviewers_no_flag = [r.strip("!") for r in reviewers]
    reviewers = [r for r in reviewers_no_flag if r.lower() not in blockers]

    # Add all blockers to reviewers list. Reviewers list will contain a list
    # of all reviewers where blocking ones are marked with the "!" suffix.
    reviewers += make_blocking(blockers)
    reviewers = remove_duplicates(reviewers)
    lowercase_reviewers = [r.lower() for r in reviewers]
    lowercase_blockers = [r.lower() for r in blockers]

    for commit in commits:
        if reviewers:
            # Only the reviewers mentioned in command line args will remain.
            # New reviewers will be marked as "granted" (r=).
            granted = reviewers.copy()
            requested = []

            # commit["reviewers"]["request|"] is a list containing reviewers
            # provided in commit's title with an r? mark.
            for reviewer in commit["reviewers"]["request"]:
                r = reviewer.strip("!")
                # check which request reviewers are provided also via -r
                if r.lower() in lowercase_reviewers:
                    requested.append(r)
                    granted.remove(r)
                # check which request reviewers are provided also via -R
                elif r.lower() in lowercase_blockers:
                    requested.append("%s!" % r)
                    granted.remove("%s!" % r)
        else:
            granted = commit["reviewers"].get("granted", [])
            requested = commit["reviewers"].get("request", [])
        commit["reviewers"] = dict(granted=granted, request=requested)
        commit["has-reviewers"] = bool(granted + requested)

        if args.bug:
            # Bug ID command arg used.
            commit["bug-id"] = args.bug

        # Mark a commit as WIP if --wip is provided, or if the commit does not have
        # any reviewers.  This is in addition to checking for the WIP: prefix in
        # helpers.augment_commits_from_body()
        if not args.no_wip and (args.wip or not commit["has-reviewers"]):
            commit["wip"] = True
        else:
            commit.setdefault("wip", False)

    # Honour config setting to always use blockers
    if not reviewers and config.always_blocking:
        for commit in commits:
            commit["reviewers"] = dict(
                request=make_blocking(commit["reviewers"]["request"]),
                granted=make_blocking(commit["reviewers"]["granted"]),
            )

    if args.command == "uplift":
        update_commits_for_uplift(commits)

    update_commit_title_previews(commits)


def update_commits_for_uplift(commits):
    """Prepares a set of commits for uplifting."""
    for commit in commits:
        # When uplifting, ensure the `Differential Revision` line is properly
        # moved, and that `rev-id` is updated to not point at the original revision.
        commit["body"], commit["rev-id"] = move_drev_to_original(
            commit["body"],
            commit["rev-id"],
        )

        # Clear all reviewers from the revision.
        commit["reviewers"] = {
            "granted": [],
            "request": [],
        }


def update_revision_description(transactions, commit, revision):
    # Appends differential.revision.edit transaction(s) to `transactions` if
    # updating the commit title and/or summary is required.

    if commit["title"] != revision["fields"]["title"]:
        transactions.append(dict(type="title", value=commit["title"]))

    # The Phabricator API will refuse the new summary value if we include the
    # "Differential Revision:" keyword in the summary body.
    local_body = strip_differential_revision(commit["body"]).strip()
    remote_body = strip_differential_revision(revision["fields"]["summary"]).strip()
    if local_body != remote_body:
        transactions.append(dict(type="summary", value=local_body))


def update_revision_bug_id(transactions, commit, revision):
    # Appends differential.revision.edit transaction(s) to `transactions` if
    # updating the commit bug-id is required.
    if commit["bug-id"] and commit["bug-id"] != revision["fields"]["bugzilla.bug-id"]:
        transactions.append(dict(type="bugzilla.bug-id", value=commit["bug-id"]))


def local_uplift_if_possible(args, repo, commits) -> bool:
    """If possible, rebase local repository commits onto the target uplift train.

    Uplifts will be performed if the `--no-rebase` argument is not
    passed, if the local repository has a corresponding unified head
    for the uplift train, and if the revset being submitted is not
    already descendant from the target unified head.

    Returns a `bool` indicating if the `commits` should avoid making local
    repository changes to reflect new Phabricator revisions. If `True`,
    local commits may be amended to reflect their state on Phabricator,
    for example to add the `Differential Revision` line. If `False`, the
    local repository commits should not be updated by `moz-phab`.
    """
    if args.no_rebase:
        # If args tell us not to do a rebase, do not make any local changes and
        # return without rebasing. This is the same as submitting an uplift where
        # the original patch is sent to Phabricator without any modifications.
        # In this case we want to avoid local amendments to commits.
        return True

    # Try and find a local repo identifier (hg bookmark, git remote branch) to rebase
    # our revset onto.
    unified_head = repo.map_callsign_to_unified_head(args.train)

    if not unified_head:
        # If we didn't find a unified head, we intend to submit an uplift without
        # modifying the local repo state via a rebase.
        return True

    if not repo.is_descendant(unified_head):
        # If we found a head to rebase onto and the commit isn't already a descendant
        # of our target identifier, uplift the commits onto the unified head.
        with wait_message(f"Rebasing commits onto {unified_head}"):
            commits = repo.uplift_commits(unified_head, commits)

    return False


def _submit(repo, args):
    telemetry().submission.preparation_time.start()
    with wait_message("Checking connection to Phabricator."):
        # Check if raw Conduit API can be used
        if not conduit.check():
            raise Error("Failed to use Conduit API")

        # Check if local and remote VCS matches
        repo.check_vcs()

    repo.before_submit()

    # Find and preview commits to submits.
    with wait_message("Looking for commits.."):
        commits = repo.commit_stack(single=args.single)
    if not commits:
        raise Error("Failed to find any commits to submit")

    if args.command == "uplift":
        # Perform uplift logic during submission.
        avoid_local_changes = local_uplift_if_possible(args, repo, commits)
    else:
        avoid_local_changes = False

    with wait_message("Loading commits.."):
        # Pre-process to load metadata.
        morph_blocking_reviewers(commits)
        augment_commits_from_body(commits)
        update_commits_from_args(commits, args)

    # Display a one-line summary of commit and WIP count.
    commit_count = len(commits)
    wip_commit_count = sum(1 for commit in commits if commit["wip"])

    if wip_commit_count == commit_count:
        status = "as Work In Progress"
    elif wip_commit_count:
        status = f"{wip_commit_count} as Work In Progress"
    else:
        status = "for review"

    logger.warning(f"Submitting {commit_count} commit{'s'[:commit_count^1]} {status}")

    # Validate commit stack is suitable for review.
    show_commit_stack(commits, validate=True)
    try:
        with wait_message("Checking commits.."):
            repo.check_commits_for_submit(commits, require_bug=not args.no_bug)
    except Error as e:
        if not args.force:
            raise Error("Unable to submit commits:\n\n%s" % e)
        logger.error("Ignoring issues found with commits:\n\n%s", e)

    if not any(commit["submit"] for commit in commits):
        logger.warning("No changes to submit.")
        return

    # Show a warning if there are untracked files.
    if config.warn_untracked:
        untracked = repo.untracked()
        if untracked:
            logger.warning(
                "Warning: found %s untracked file%s (will not be submitted):",
                len(untracked),
                "" if len(untracked) == 1 else "s",
            )
            if len(untracked) <= 5:
                for filename in untracked:
                    logger.info("  %s", filename)

    # Show a warning if -m is used and there are new commits.
    if args.message and any([c for c in commits if not c["rev-id"]]):
        logger.warning(
            "Warning: --message works with updates only, and will not\n"
            "result in a comment on new revisions."
        )

    telemetry().submission.preparation_time.stop()
    telemetry().submission.commits_count.add(len(commits))

    # Confirmation prompt.
    if args.yes:
        pass
    elif config.auto_submit and not args.interactive:
        logger.info(
            "Automatically submitting (as per submit.auto_submit in %s)", config.name
        )
    else:
        res = prompt(
            "Submit to %s" % PHABRICATOR_URLS.get(repo.phab_url, repo.phab_url),
            ["Yes", "No", "Always"],
        )
        if res == "No":
            return
        if res == "Always":
            config.auto_submit = True
            config.write()

    # Process.
    telemetry().submission.process_time.start()
    previous_commit = None
    # Collect all existing revisions to get reviewers info.
    rev_ids_to_update = [int(c["rev-id"]) for c in commits if c.get("rev-id")]
    revisions_to_update = None
    if rev_ids_to_update:
        with wait_message("Loading revision data..."):
            list_to_update = conduit.get_revisions(ids=rev_ids_to_update)

        revisions_to_update = {str(r["id"]): r for r in list_to_update}

    last_node = commits[-1]["orig-node"]
    for commit in commits:
        diff = None

        check_in_needed = args.check_in_needed and commit["orig-node"] == last_node
        # Only revisions being updated have an ID.  Newly created ones don't.
        if not commit["submit"]:
            previous_commit = commit
            continue

        is_update = bool(commit["rev-id"])
        revision_to_update = (
            revisions_to_update[commit["rev-id"]] if is_update else None
        )
        has_existing_reviewers = (
            bool(revision_to_update["attachments"]["reviewers"]["reviewers"])
            if revision_to_update
            else False
        )

        # Let the user know something's happening.
        if is_update:
            logger.info("\nUpdating revision D%s:", commit["rev-id"])
        else:
            logger.info("\nCreating new revision:")

        logger.info("%s %s", commit["name"], revision_title_from_commit(commit))
        repo.checkout(commit["node"])

        # WIP submissions shouldn't set reviewers on phabricator.
        if commit["wip"]:
            reviewers = ""
        else:
            reviewers = ", ".join(
                commit["reviewers"]["granted"] + commit["reviewers"]["request"]
            )

        # Create arc-annotated commit description.
        template_vars = dict(
            title=revision_title_from_commit(commit),
            body=commit["body"],
            reviewers=reviewers,
            bug_id=commit["bug-id"],
        )
        summary = commit["body"]
        if previous_commit and not args.no_stack:
            template_vars["depends_on"] = "Depends on D%s" % previous_commit["rev-id"]
            summary = "%s\n\n%s" % (summary, template_vars["depends_on"])

        message = arc_message(template_vars)

        # Create a diff if needed
        with wait_message("Creating local diff..."):
            diff = repo.get_diff(commit)

        if diff:
            telemetry().submission.files_count.add(len(diff.changes))
            with wait_message("Uploading binary file(s)..."):
                diff.upload_files()

            with wait_message("Submitting the diff..."):
                diff.submit(commit, message)

        if is_update:
            with wait_message("Updating revision..."):
                rev = conduit.update_revision(
                    commit,
                    has_existing_reviewers,
                    diff_phid=diff.phid,
                    comment=args.message,
                    check_in_needed=check_in_needed,
                )
        else:
            with wait_message("Creating a new revision..."):
                rev = conduit.create_revision(
                    commit,
                    summary,
                    diff.phid,
                    check_in_needed=check_in_needed,
                )

        revision_url = "%s/D%s" % (repo.phab_url, rev["object"]["id"])

        # Append/replace div rev url to/in commit description.
        body = amend_revision_url(commit["body"], revision_url)

        # Amend the commit if required.
        # As commit rewriting can be expensive we avoid it in some circumstances, such
        # as pre-pending "WIP: " to commits submitted as WIP to Phabricator.
        if commit["title-preview"] != commit["title"] or body != commit["body"]:
            commit["title"] = commit["title-preview"]
            commit["body"] = body
            commit["rev-id"] = parse_arc_diff_rev(commit["body"])

            if not avoid_local_changes:
                with wait_message("Updating commit.."):
                    repo.amend_commit(commit, commits)

        # Diff property has to be set after potential SHA1 change.
        if diff:
            with wait_message("Setting diff metadata..."):
                diff.set_property(commit, message)

        previous_commit = commit

    # Cleanup (eg. strip nodes) and refresh to ensure the stack is right for the
    # final showing.
    with wait_message("Cleaning up.."):
        repo.finalize(commits)
        repo.after_submit()
        repo.cleanup()
        repo.refresh_commit_stack(commits)

    logger.warning("\nCompleted")
    show_commit_stack(
        commits, validate=False, show_rev_urls=True, show_updated_only=True
    )
    telemetry().submission.process_time.stop()


def submit(repo, args):
    try:
        _submit(repo, args)
    except Exception as e:
        # Check if the `fallback` flag was set and print a warning message.
        if hasattr(args, "fallback") and args.fallback:
            logger.warning(
                "You didn't specify a valid command, so we ran `submit` for you, "
                "and it failed."
            )

        raise e


def add_parser(parser):
    submit_parser = parser.add_parser(
        "submit",
        aliases=["upload"],
        help="Submit commit(s) to Phabricator",
        description=(
            "MozPhab will change the working directory and amend the commits during "
            "the submission process."
        ),
    )
    add_submit_arguments(submit_parser)
    submit_parser.set_defaults(func=submit, needs_repo=True)


def add_submit_arguments(parser):
    """Add `moz-phab submit` command line arguments to a parser."""
    parser.add_argument(
        "--path", "-p", help="Set path to repository (default: detected)."
    )
    parser.add_argument(
        "--yes",
        "-y",
        action="store_true",
        help="Submit without confirmation (default: %s)." % config.auto_submit,
    )
    parser.add_argument(
        "--interactive",
        "-i",
        action="store_true",
        help="Submit with confirmation (default: %s)." % (not config.auto_submit),
    )
    parser.add_argument(
        "--message",
        "-m",
        help="Provide a custom update message (default: none).",
    )
    parser.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Override sanity checks and force submission; a tool of last resort.",
    )
    parser.add_argument(
        "--force-delete",
        action="store_true",
        help="Mercurial only: Ignore error caused by a DAG branch point without "
        "evolve installed.",
    )
    parser.add_argument(
        "--bug", "-b", help="Set Bug ID for all commits (default: from commit)."
    )
    parser.add_argument(
        "--no-bug",
        action="store_true",
        help="Continue if a bug number is not provided.",
    )
    parser.add_argument(
        "--reviewer",
        "--reviewers",
        "-r",
        action="append",
        help="Set review(s) for all commits (default: from commit).",
    )
    parser.add_argument(
        "--blocker",
        "--blockers",
        "-R",
        action="append",
        help="Set blocking review(s) for all commits (default: from commit).",
    )
    parser.add_argument(
        "--nolint",
        "--no-lint",
        action="store_true",
        help="Do not run lint (default: lint changed files if configured).",
    )
    parser.add_argument(
        "--check-in-needed",
        action="store_true",
        help="Add a `check-in-needed tag to the top most revision.",
    )
    wip_group = parser.add_mutually_exclusive_group()
    wip_group.add_argument(
        "--wip",
        "--plan-changes",
        action="store_true",
        help="Create or update a revision without requesting a code review.",
    )
    wip_group.add_argument(
        "--no-wip",
        action="store_true",
        help="Don't mark reviewer-less commits as work-in-progress. Commits "
        "with descriptions that start with WIP: will continue to be "
        "flagged as work-in-progress.",
    )
    parser.add_argument(
        "--less-context",
        "--lesscontext",
        action="store_true",
        dest="lesscontext",
        help=(
            "Normally, files are diffed with full context: the entire file "
            "is sent to Phabricator so reviewers can 'show more' and see "
            "it. If you are making changes to very large files with tens of "
            "thousands of lines, this may not work well. With this flag, a "
            "revision will be created that has only a few lines of context."
        ),
    )
    parser.add_argument(
        "--no-stack",
        action="store_true",
        help="Submit multiple commits, but do not mark them as dependent.",
    )
    parser.add_argument(
        "--upstream",
        "--remote",
        "-u",
        action="append",
        help='Set upstream branch to detect the starting commit (default: "").',
    )
    parser.add_argument(
        "--force-vcs",
        action="store_true",
        help="EXPERIMENTAL: Override VCS compatibility check.",
    )
    parser.add_argument(
        "--safe-mode",
        dest="safe_mode",
        action="store_true",
        help="Run VCS with only necessary extensions.",
    )
    parser.add_argument(
        "--single",
        "-s",
        action="store_true",
        help="Submit a single commit.",
    )
    parser.add_argument(
        "start_rev",
        nargs="?",
        default=environment.DEFAULT_START_REV,
        help="Start revision of range to submit (default: detected).",
    )
    parser.add_argument(
        "end_rev",
        nargs="?",
        default=environment.DEFAULT_END_REV,
        help="End revision of range to submit (default: current commit).",
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import hashlib
import json
import time
import uuid

from typing import Optional

from mozphab import environment
from pathlib import Path

from .bmo import bmo
from .conduit import conduit, ConduitAPIError
from .logger import logger


USER_INFO_FILE = Path(environment.MOZBUILD_PATH) / "user_info.json"
EMPLOYEE_CHECK_FREQUENCY = 24 * 7 * 60 * 60  # week
MOZILLA_EMPLOYEE_EMAIL_ENDINGS = {
    "@getpocket.com",
    "@mozilla.com",
    "@mozillafoundation.org",
}


def is_bad_uuid(key: str, value: Optional[str]) -> bool:
    """Return `True` if the key/value pair corresponds to a faulty UUID."""
    return (
        key in {"installation_id", "user_code"}
        and value is not None
        and len(value) == 32
    )


def format_uuid(bad_uuid: str) -> str:
    """Turn a non-hyphenated UUID `str` into a hyphenated one.

    This function assumes the `bad_uuid` is a 32-character `str`,
    usually already having been passed through `is_bad_uuid`.

    See https://bugzilla.mozilla.org/show_bug.cgi?id=1788719 for more.
    """
    # Pass the 32-char string into a `UUID` object, then pass through
    # `str()` to get the hyphenated 36-character version.
    return str(uuid.UUID(bad_uuid))


class UserData:
    is_employee = None
    user_code = None
    installation_id = None
    last_check = None
    keys = ["is_employee", "user_code", "installation_id", "last_check"]

    def __init__(self):
        self.set_from_file()

    @property
    def is_data_collected(self):
        """True if all user info data are collected."""
        # All values are set (not None)
        return None not in [getattr(self, k) for k in self.keys]

    def to_dict(self):
        return {k: getattr(self, k) for k in self.keys}

    def update_from_dict(self, dictionary):
        """Assign attributes from a dict."""
        for key in self.keys:
            if key in dictionary:

                # See bug 1788719.
                if is_bad_uuid(key, dictionary[key]):
                    dictionary[key] = format_uuid(dictionary[key])

                setattr(self, key, dictionary[key])

    def set_from_file(self):
        """Read user info from file."""
        if USER_INFO_FILE.exists():
            with USER_INFO_FILE.open("r", encoding="utf-8") as f:
                user_info = json.load(f)
                self.update_from_dict(user_info)

    def save_user_info(self, **kwargs):
        """Save any fields provided as kwargs into the user_info file."""
        self.update_from_dict(kwargs)
        user_info = self.to_dict()
        with USER_INFO_FILE.open("w", encoding="utf-8") as f:
            json.dump(user_info, f, sort_keys=True, indent=2)

    def whoami(self):
        """Returns a dict with email and employee status."""
        # Check user in Phabricator.
        try:
            who = conduit.whoami()
        except ConduitAPIError as e:
            logger.error(str(e))
            return None

        response = dict(email=who["primaryEmail"])

        if not response["email"]:
            # If `primaryEmail` is empty we log a warning and return.
            logger.warning(
                "You have not set a primary email address in Phabricator.\n"
                "Please set a primary email address in your Phabricator settings."
            )
            response["is_employee"] = False
            return response

        lower_email = response["email"].lower()
        if any(
            lower_email.endswith(domain) for domain in MOZILLA_EMPLOYEE_EMAIL_ENDINGS
        ):
            response["is_employee"] = True
            return response

        bmo_who = bmo.whoami()
        response["is_employee"] = (
            bmo_who is not None and "mozilla-employee-confidential" in bmo_who["groups"]
        )
        return response

    def set_user_data(self, from_file_only=False):
        """Sets user data if needed.

        Returns a bool value indicating if status is updated.
        """
        if USER_INFO_FILE.exists():
            update = (
                self.last_check is None
                or time.time() - self.last_check > EMPLOYEE_CHECK_FREQUENCY
            )
            if not update and self.is_data_collected:
                return False

        if from_file_only:
            return False

        whoami = self.whoami()
        if whoami is None or not whoami["email"]:
            # `user.whoami` failed.
            return False

        is_employee = self.is_employee

        if self.installation_id is None:
            self.installation_id = str(uuid.uuid4())

        user_code = hashlib.md5(whoami["email"].encode("utf-8")).hexdigest()
        self.user_code = format_uuid(user_code)

        self.last_check = int(time.time())
        self.is_employee = whoami["is_employee"]
        self.save_user_info(
            is_employee=self.is_employee,
            user_code=self.user_code,
            installation_id=self.installation_id,
            last_check=self.last_check,
        )
        return is_employee != self.is_employee


user_data = UserData()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from unittest import mock
from unittest.mock import MagicMock

from mozphab import mozphab
from mozphab.config import config


@mock.patch("mozphab.mozphab.parse_args")
@mock.patch("mozphab.mozphab.init_sentry")
@mock.patch("mozphab.mozphab.check_for_updates", new=mock.Mock(return_value=None))
def test_sentry_not_enabled_if_development(mock_init_sentry, mock_parse_args):
    config.report_to_sentry = True
    args = MagicMock()
    args.needs_repo = False  # skip FS operations
    mock_parse_args.return_value = args

    mozphab.main([], is_development=True)

    mock_init_sentry.assert_not_called()


@mock.patch("mozphab.mozphab.parse_args")
@mock.patch("mozphab.mozphab.init_sentry")
@mock.patch("mozphab.mozphab.check_for_updates", new=mock.Mock(return_value=None))
def test_sentry_not_enabled_if_config_disabled(mock_init_sentry, mock_parse_args):
    config.report_to_sentry = False
    args = MagicMock()
    args.needs_repo = False  # skip FS operations
    mock_parse_args.return_value = args

    mozphab.main([], is_development=True)

    mock_init_sentry.assert_not_called()


@mock.patch("mozphab.mozphab.parse_args")
@mock.patch("mozphab.mozphab.init_sentry")
@mock.patch("mozphab.mozphab.telemetry")
@mock.patch("mozphab.mozphab.check_for_updates", new=mock.Mock(return_value=None))
def test_sentry_enabled(_telemetry, mock_init_sentry, mock_parse_args):
    config.report_to_sentry = True
    args = MagicMock()
    args.needs_repo = False  # skip FS operations
    mock_parse_args.return_value = args

    mozphab.main([], is_development=False)

    mock_init_sentry.assert_called_once()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozphab.commands.submit import (
    update_commits_for_uplift,
    local_uplift_if_possible,
)


def test_local_uplift_if_possible():
    class Args:
        def __init__(self, no_rebase=False, train="train"):
            self.no_rebase = no_rebase
            self.train = train

    class Repo:
        def __init__(self, unified_head="beta", is_descendant=True):
            self.unified_head = unified_head
            self._is_descendant = is_descendant
            self.uplift_called = False

        def map_callsign_to_unified_head(self, *args, **kwargs):
            return self.unified_head

        def is_descendant(self, *args, **kwargs):
            return self._is_descendant

        def uplift_commits(self, *args, **kwargs):
            self.uplift_called = True

    commits = [
        {
            "title": "A",
            "reviewers": dict(granted=["john"], request=[]),
            "bug-id": None,
            "body": "",
            "rev-id": 1,
        },
    ]

    repo = Repo()

    args = Args(no_rebase=True)
    assert (
        local_uplift_if_possible(args, repo, commits) is True
    ), "Should always do a one-off uplift when `--no-rebase` is set."

    args = Args()
    repo = Repo(unified_head=None)

    assert (
        local_uplift_if_possible(args, repo, commits) is True
    ), "Should avoid do a one-off when no unified head is found."

    repo = Repo(is_descendant=True)
    assert (
        local_uplift_if_possible(args, repo, commits) is False
    ), "Should avoid uplifting commits locally when destination is a descendant."

    # Rebase-uplift case.
    repo = Repo(
        is_descendant=False,
        unified_head="beta",
    )
    args = Args(
        no_rebase=False,
        train="beta",
    )
    assert (
        local_uplift_if_possible(args, repo, commits) is False
    ), "Uplifting commits locally should amend them as well."
    assert (
        repo.uplift_called
    ), "Should call `uplift_commits` when non-descendant unified head found."


def test_update_commits_for_uplift_sets_relman_review():
    commits = [
        {
            "title": "A",
            "reviewers": dict(granted=["john"], request=[]),
            "bug-id": None,
            "body": "",
            "rev-id": 1,
        },
    ]

    update_commits_for_uplift(commits)

    reviewers = commits[0]["reviewers"]

    assert not reviewers[
        "request"
    ], "Uplifted patch should have no requested reviewers initially."

    assert not reviewers[
        "granted"
    ], "Uplifted patch should have no granted reviewers initially."


def test_update_commits_for_uplift_sets_original_revision():
    commits = [
        # Check initial submission.
        {
            "title": "bug 1: firstline r?reviewer",
            "reviewers": dict(granted=["john"], request=[]),
            "bug-id": 1,
            "body": (
                "bug 1: firstline r?reviewer\n"
                "\n"
                "Differential Revision: https://phabricator.services.mozila.com/D1\n"
            ),
            "rev-id": 1,
        },
        # Check update of existing uplift revision.
        {
            "title": "bug 1: firstline r?reviewer",
            "reviewers": dict(granted=["john"], request=[]),
            "bug-id": 1,
            "body": (
                "bug 1: firstline r?reviewer\n"
                "\n"
                "Original Revision: https://phabricator.services.mozila.com/D1\n"
                "\n"
                "Differential Revision: https://phabricator.services.mozila.com/D2\n"
            ),
            "rev-id": 2,
        },
    ]

    update_commits_for_uplift(commits)

    # Initial submission.
    body = commits[0]["body"]
    rev_id = commits[0]["rev-id"]

    assert "Differential Revision:" not in body
    assert "Original Revision:" in body
    assert rev_id is None

    # Update of existing uplift.
    body = commits[1]["body"]
    rev_id = commits[1]["rev-id"]

    assert "Differential Revision:" in body
    assert "Original Revision:" in body
    assert rev_id == 2



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json
import pytest

from unittest import mock

from mozphab.bmo import bmo, BMOAPIError
from mozphab.exceptions import Error


@mock.patch("mozphab.bmo.BMOAPI._req_with_retries")
@mock.patch("mozphab.bmo.conduit")
def test_whoami(m_conduit, m_req_with_retries):
    bmo.whoami()
    m_req_with_retries.assert_called_once_with(
        "whoami", headers={"X-PHABRICATOR-TOKEN": mock.ANY}
    )
    m_conduit.load_api_token.assert_called_once()


@mock.patch("mozphab.bmo.conduit")
def test_build_request(m_conduit):
    m_conduit.repo.bmo_url = "https://bmo.test"

    assert bmo._build_request(method="test_method") == {
        "url": "https://bmo.test/rest/test_method",
        "method": "GET",
        "headers": {"User-Agent": mock.ANY},
    }

    assert bmo._build_request(method="test_method", headers={"X-Test": "true"}) == {
        "url": "https://bmo.test/rest/test_method",
        "method": "GET",
        "headers": {"User-Agent": mock.ANY, "X-Test": "true"},
    }


def test_sanitised_req():
    assert bmo._sanitise_req(
        {
            "url": "https://bmo.test/rest/test_method",
            "method": "GET",
            "headers": {"X-PHABRICATOR-TOKEN": "cli-secret"},
        }
    ) == {
        "url": "https://bmo.test/rest/test_method",
        "method": "GET",
        "headers": {"X-PHABRICATOR-TOKEN": "cli-XXXX"},
    }


@mock.patch("urllib.request.urlopen")
@mock.patch("mozphab.bmo.conduit")
def test_get(m_conduit, m_urlopen):
    m_conduit.repo.bmo_url = "https://bmo.test"

    # build fake context-manager to mock urlopen
    cm = mock.MagicMock()
    cm.getcode.return_value = 200
    cm.__enter__.return_value = cm
    m_urlopen.return_value = cm

    # success
    cm.read.return_value = json.dumps({"result": "result"})
    assert bmo.get("method") == {"result": "result"}

    # error
    cm.read.return_value = json.dumps({"error": "aieee"})
    with pytest.raises(BMOAPIError) as bmo_error:
        bmo.get("method")
    assert bmo_error.value.args[0].startswith("Bugzilla Error: ")


@mock.patch("mozphab.bmo.BMOAPI.get")
def test_req_with_retries(m_get):
    # raises Error after 3 retries
    m_get.side_effect = (BMOAPIError, BMOAPIError, BMOAPIError)
    with pytest.raises(Error):
        bmo._req_with_retries("test")

    # raises error with custom retry amount
    m_get.side_effect = (BMOAPIError, BMOAPIError)
    with pytest.raises(Error):
        bmo._req_with_retries("test", retries=2)

    # returns result if successful
    m_get.side_effect = None
    m_get.return_value = {"message": "test"}
    assert bmo._req_with_retries("test")["message"] == "test"



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os
import platform
import pytest
import shutil

from unittest import mock

from callee import Contains
from .conftest import hg_out, write_text

from mozphab import environment, mozphab

# Fail if arc ping is called
arc_ping = mock.Mock()
arc_ping.return_value = False

call_conduit = mock.Mock()
call_conduit.side_effect = ({}, [{"userName": "alice", "phid": "PHID-USER-1"}])


def test_submit_create(in_process, hg_repo_path):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        [dict(userName="alice", phid="PHID-USER-1")],
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    write_text(hg_repo_path / "A to rename", "rename me\nsecond line\n")
    write_text(hg_repo_path / "B to remove", "remove me\n")
    write_text(hg_repo_path / "C to modify", "modify me\n")
    write_text(hg_repo_path / "D to copy", "copy me\n")
    hg_out("add")
    hg_out("commit", "-m", "first")
    (hg_repo_path / "subdir").mkdir()
    hg_out("copy", "D to copy", "D copied")
    write_text(hg_repo_path / "subdir" / "E add", "added\n")
    (hg_repo_path / "A to rename").rename(hg_repo_path / "A renamed")
    (hg_repo_path / "B to remove").unlink()
    write_text(hg_repo_path / "C to modify", "modified\n")
    hg_out("addremove")
    write_text(hg_repo_path / "msg", "Ą r?alice", encoding="utf-8")
    hg_out("commit", "-l", "msg")
    mozphab.main(["submit", "--yes", "--bug", "1", "."], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - Ą r?alice

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert mock.call("conduit.ping", {}) in call_conduit.call_args_list
    assert (
        mock.call("user.query", dict(usernames=["alice"]))
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "diffusion.repository.search",
            dict(limit=1, constraints=dict(callsigns=["TEST"])),
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.setdiffproperty",
            {"diff_id": "1", "name": "local:commits", "data": Contains('"rev": "')},
        )
        in call_conduit.call_args_list
    )
    assert (
        call_conduit.call_args_list.count(
            mock.call("differential.setdiffproperty", mock.ANY)
        )
        == 2
    )
    assert (
        mock.call(
            "differential.creatediff",
            {
                "sourceControlPath": "/",
                "sourceControlSystem": "hg",
                "lintStatus": "none",
                "sourcePath": mock.ANY,
                "unitStatus": "none",
                "sourceMachine": "http://example.test",
                "sourceControlBaseRevision": mock.ANY,
                "repositoryPHID": "PHID-REPO-1",
                "branch": "default",
                "changes": [
                    {
                        "currentPath": "A renamed",
                        "type": 6,  # MOVE_HERE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 2,
                                "newOffset": 1,
                                "newLength": 2,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " rename me\n second line\n",
                            }
                        ],
                        "oldProperties": {},
                        "oldPath": "A to rename",
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "metadata": {},
                        "newProperties": {},
                        "fileType": 1,
                    },
                    {
                        "currentPath": "A to rename",
                        "type": 4,  # MOVE_AWAY
                        "hunks": [],
                        "oldProperties": {},
                        "oldPath": None,
                        "commitHash": mock.ANY,
                        "awayPaths": ["A renamed"],
                        "metadata": {},
                        "newProperties": {},
                        "fileType": 1,
                    },
                    {
                        "currentPath": "B to remove",
                        "type": 3,  # DELETE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 0,
                                "newLength": 0,
                                "addLines": 0,
                                "delLines": 1,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": "-remove me\n",
                            }
                        ],
                        "awayPaths": [],
                        "fileType": 1,
                        "oldPath": "B to remove",
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "metadata": {},
                        "oldProperties": {"unix:filemode": "100644"},
                    },
                    {
                        "currentPath": "C to modify",
                        "type": 2,  # CHANGE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 1,
                                "delLines": 1,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": "-modify me\n+modified\n",
                            }
                        ],
                        "commitHash": mock.ANY,
                        "metadata": {},
                        "fileType": 1,
                        "oldPath": "C to modify",
                        "newProperties": {},
                        "awayPaths": [],
                        "oldProperties": {},
                    },
                    {
                        "metadata": {},
                        "oldPath": "D to copy",
                        "currentPath": "D copied",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 7,  # COPY HERE
                        "fileType": 1,
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " copy me\n",
                            }
                        ],
                    },
                    {
                        "metadata": {},
                        "oldPath": None,
                        "currentPath": "D to copy",
                        "awayPaths": ["D copied"],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 5,  # COPY AWAY
                        "fileType": 1,
                        "hunks": [],
                    },
                    {
                        "currentPath": "subdir/E add",
                        "type": 1,  # ADD
                        "hunks": [
                            {
                                "corpus": "+added\n",
                                "addLines": 1,
                                "oldOffset": 0,
                                "newOffset": 1,
                                "newLength": 1,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "oldLength": 0,
                                "isMissingNewNewline": False,
                            }
                        ],
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "newProperties": {"unix:filemode": "100644"},
                        "oldPath": None,
                        "oldProperties": {},
                        "fileType": 1,
                        "metadata": {},
                    },
                ],
                "creationMethod": "moz-phab-hg",
            },
        )
        in call_conduit.call_args_list
    )


def test_submit_create_no_trailing_newline(in_process, hg_repo_path):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        [dict(userName="alice", phid="PHID-USER-1")],
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    write_text(hg_repo_path / "A to rename", "rename me\nsecond line")
    write_text(hg_repo_path / "B to remove", "remove me")
    write_text(hg_repo_path / "C to modify", "modify me")
    write_text(hg_repo_path / "D to copy", "copy me")
    hg_out("add")
    hg_out("commit", "-m", "first")
    (hg_repo_path / "subdir").mkdir()
    hg_out("copy", "D to copy", "D copied")
    write_text(hg_repo_path / "subdir" / "E add", "added")
    (hg_repo_path / "A to rename").rename(hg_repo_path / "A renamed")
    (hg_repo_path / "B to remove").unlink()
    write_text(hg_repo_path / "C to modify", "modified")
    hg_out("addremove")
    write_text(hg_repo_path / "msg", "Ą r?alice", encoding="utf-8")
    hg_out("commit", "-l", "msg")
    mozphab.main(["submit", "--yes", "--bug", "1", "."], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - Ą r?alice

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert mock.call("conduit.ping", {}) in call_conduit.call_args_list
    assert (
        mock.call("user.query", dict(usernames=["alice"]))
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "diffusion.repository.search",
            dict(limit=1, constraints=dict(callsigns=["TEST"])),
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.setdiffproperty",
            {"diff_id": "1", "name": "local:commits", "data": Contains('"rev": "')},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.creatediff",
            {
                "sourceControlPath": "/",
                "sourceControlSystem": "hg",
                "lintStatus": "none",
                "sourcePath": mock.ANY,
                "unitStatus": "none",
                "sourceMachine": "http://example.test",
                "sourceControlBaseRevision": mock.ANY,
                "repositoryPHID": "PHID-REPO-1",
                "branch": "default",
                "changes": [
                    {
                        "currentPath": "A renamed",
                        "type": 6,  # MOVE_HERE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 2,
                                "newOffset": 1,
                                "newLength": 2,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " rename me\n second line",
                            }
                        ],
                        "oldProperties": {},
                        "oldPath": "A to rename",
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "metadata": {},
                        "newProperties": {},
                        "fileType": 1,
                    },
                    {
                        "currentPath": "A to rename",
                        "type": 4,  # MOVE_AWAY
                        "hunks": [],
                        "oldProperties": {},
                        "oldPath": None,
                        "commitHash": mock.ANY,
                        "awayPaths": ["A renamed"],
                        "metadata": {},
                        "newProperties": {},
                        "fileType": 1,
                    },
                    {
                        "currentPath": "B to remove",
                        "type": 3,  # DELETE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 2,
                                "newOffset": 0,
                                "newLength": 0,
                                "addLines": 0,
                                "delLines": 1,
                                "isMissingOldNewline": True,
                                "isMissingNewNewline": False,
                                "corpus": "-remove me\n\\ No newline at end of file\n",
                            }
                        ],
                        "awayPaths": [],
                        "fileType": 1,
                        "oldPath": "B to remove",
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "metadata": {},
                        "oldProperties": {"unix:filemode": "100644"},
                    },
                    {
                        "currentPath": "C to modify",
                        "type": 2,  # CHANGE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 1,
                                "delLines": 1,
                                "isMissingOldNewline": True,
                                "isMissingNewNewline": True,
                                "corpus": (
                                    "-modify me\n"
                                    "\\ No newline at end of file\n"
                                    "+modified\n"
                                    "\\ No newline at end of file\n"
                                ),
                            }
                        ],
                        "commitHash": mock.ANY,
                        "metadata": {},
                        "fileType": 1,
                        "oldPath": "C to modify",
                        "newProperties": {},
                        "awayPaths": [],
                        "oldProperties": {},
                    },
                    {
                        "metadata": {},
                        "oldPath": "D to copy",
                        "currentPath": "D copied",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 7,  # COPY HERE
                        "fileType": 1,
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " copy me",
                            }
                        ],
                    },
                    {
                        "metadata": {},
                        "oldPath": None,
                        "currentPath": "D to copy",
                        "awayPaths": ["D copied"],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 5,  # COPY AWAY
                        "fileType": 1,
                        "hunks": [],
                    },
                    {
                        "currentPath": "subdir/E add",
                        "type": 1,  # ADD
                        "hunks": [
                            {
                                "oldOffset": 0,
                                "oldLength": 0,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 1,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": True,
                                "corpus": "+added\n\\ No newline at end of file\n",
                            }
                        ],
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "newProperties": {"unix:filemode": "100644"},
                        "oldPath": None,
                        "oldProperties": {},
                        "fileType": 1,
                        "metadata": {},
                    },
                ],
                "creationMethod": "moz-phab-hg",
            },
        )
        in call_conduit.call_args_list
    )


def test_submit_create_no_bug(in_process, hg_repo_path):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        [dict(userName="alice", phid="PHID-USER-1")],
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    write_text(hg_repo_path / "x", "a")
    hg_out("add")
    hg_out("commit", "--message", "A r?alice")
    mozphab.main(["submit", "--yes", "--no-bug", "."], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
A r?alice

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()


def test_submit_create_binary(in_process, hg_repo_path, data_file):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        # file.allocate
        dict(dict(filePHID=None, upload=True)),
        # file.upload
        dict(),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    shutil.copyfile(str(data_file), str(hg_repo_path / "img.png"))
    hg_out("add")
    hg_out("commit", "-m", "IMG")

    mozphab.main(["submit", "--yes", "--bug", "1", "."], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - IMG

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "file.allocate",
            {"name": "img.png", "contentHash": mock.ANY, "contentLength": 182},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call("file.upload", {"data_base64": mock.ANY, "name": "img.png"})
        in call_conduit.call_args_list
    )


def test_submit_create_binary_existing(in_process, hg_repo_path, data_file):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        # file.allocate
        dict(dict(filePHID="PHID-FILE-1", upload=False)),
        # no file.upload call
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    shutil.copyfile(str(data_file), str(hg_repo_path / "img.png"))
    hg_out("add")
    hg_out("commit", "-m", "IMG")

    mozphab.main(["submit", "--yes", "--bug", "1", "."], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - IMG

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "file.allocate",
            {"name": "img.png", "contentHash": mock.ANY, "contentLength": 182},
        )
        in call_conduit.call_args_list
    )
    assert mock.call("file.upload", mock.ANY) not in call_conduit.call_args_list


def test_submit_create_binary_chunked(in_process, hg_repo_path, data_file):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        # file.allocate
        dict(dict(filePHID="PHID-FILE-1", upload=True)),
        # file.querychunks
        [
            dict(byteStart="0", byteEnd="4194304", complete=False),
            dict(byteStart="4194304", byteEnd="8388608", complete=False),
            dict(byteStart="8388608", byteEnd="8425160", complete=False),
        ],
        # file.uploadchunk
        dict(),
        # file.uploadchunk
        dict(),
        # file.uploadchunk
        dict(),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    shutil.copyfile(str(data_file), str(hg_repo_path / "img.png"))
    hg_out("add")
    hg_out("commit", "-m", "IMG")

    mozphab.main(["submit", "--yes", "--bug", "1", "."], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - IMG

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "file.allocate",
            {"name": "img.png", "contentHash": mock.ANY, "contentLength": 182},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call("file.querychunks", {"filePHID": "PHID-FILE-1"})
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "file.uploadchunk",
            {
                "filePHID": "PHID-FILE-1",
                "byteStart": 0,
                "data": mock.ANY,
                "dataEncoding": "base64",
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "file.uploadchunk",
            {
                "filePHID": "PHID-FILE-1",
                "byteStart": 4194304,
                "data": mock.ANY,
                "dataEncoding": "base64",
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "file.uploadchunk",
            {
                "filePHID": "PHID-FILE-1",
                "byteStart": 8388608,
                "data": mock.ANY,
                "dataEncoding": "base64",
            },
        )
        in call_conduit.call_args_list
    )


def test_submit_remove_cr(in_process, hg_repo_path):
    if environment.IS_WINDOWS:
        pytest.skip("Removing CR will not work on Windows.")

    call_conduit.side_effect = (
        # CREATE
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        # user.search
        [dict(userName="alice", phid="PHID-USER-1")],
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
        # UPDATE
        # no need to ping (checked)
        # no need to check reviewer
        # no need to search for repository repository data is saved in .hg
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-2", diffid="2")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="124")),
        # differential.setdiffproperty
        dict(),
    )
    (hg_repo_path / "X").write_text("a\r\nb\n")
    hg_out("add")
    hg_out("commit", "--message", "A r?alice")
    mozphab.main(["submit", "--yes", "--bug", "1", "."], is_development=True)
    call_conduit.reset_mock()
    # removing CR, leaving LF
    (hg_repo_path / "X").write_text("a\nb\n")
    hg_out("commit", "--message", "B r?alice")
    mozphab.main(["submit", "--yes", "--bug", "1", "."], is_development=True)

    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 2,
                        "fileType": 1,
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 2,
                                "newOffset": 1,
                                "newLength": 2,
                                "addLines": 1,
                                "delLines": 1,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": "-a\r\n+a\n b\n",
                            }
                        ],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "hg",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-hg",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "default",
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.setdiffproperty",
            {
                "diff_id": "2",
                "name": "local:commits",
                "data": Contains('"summary": "Bug 1 - B r?alice"')
                & Contains(
                    '"message": "'
                    "Bug 1 - B r?alice\\n\\n"
                    "Summary:\\n\\n\\n\\n\\n"
                    "Test Plan:\\n\\n"
                    "Reviewers: alice\\n\\n"
                    "Subscribers:\\n\\n"
                    'Bug #: 1"'
                ),
            },
        )
        in call_conduit.call_args_list
    )


def test_submit_single_first(in_process, hg_repo_path, hg_sha):
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    write_text(hg_repo_path / "X", "a\n")
    hg_out("add", "X")
    hg_out("commit", "-m", "A")
    sha = hg_sha()
    write_text(hg_repo_path / "X", "b\n")
    hg_out("commit", "-m", "B")

    mozphab.main(
        ["submit", "--yes", "--bug", "1", "--single", sha], is_development=True
    )

    log = hg_out("log", "--template", r"{desc}\n---\n", "--limit", "2")
    expected = """\
B
---
Bug 1 - A

Differential Revision: http://example.test/D123
---
"""
    assert log == expected


def test_submit_single_last(in_process, hg_repo_path):
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    write_text(hg_repo_path / "X", "a\n")
    hg_out("add", "X")
    hg_out("commit", "-m", "A")
    write_text(hg_repo_path / "X", "b\n")
    hg_out("commit", "-m", "B")

    mozphab.main(["submit", "--yes", "--bug", "1", "--single"], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n---\n", "--limit", "2")
    expected = """\
Bug 1 - B

Differential Revision: http://example.test/D123
---
A
---
"""
    assert log == expected


def test_multiple_copy(in_process, hg_repo_path):
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    write_text(hg_repo_path / "X", "a\n")
    hg_out("add", "X")
    hg_out("commit", "-m", "A")
    hg_out("cp", "X", "X_copy_1")
    hg_out("cp", "X", "X_copy_2")
    hg_out("commit", "-m", "multiple copy")

    mozphab.main(["submit", "--yes", "--bug", "1", "--single"], is_development=True)

    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": None,
                        "currentPath": "X",
                        "awayPaths": ["X_copy_1", "X_copy_2"],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 5,  # COPY_AWAY
                        "fileType": 1,
                        "hunks": [],
                    },
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X_copy_1",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 7,
                        "fileType": 1,
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " a\n",
                            }
                        ],
                    },
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X_copy_2",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 7,
                        "fileType": 1,
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " a\n",
                            }
                        ],
                    },
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "hg",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-hg",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "default",
            },
        )
        in call_conduit.call_args_list
    )


# To re-enable this test for Windows we would need a way to modify
# UMASK bits similar to how it's done in test_integration_git.py.
# However, Mercurial doesn't have an equivalent to `git update-index`.
@pytest.mark.skipif(
    platform.system() == "Windows",
    reason="Skipped because Windows cannot modify UMASK bits",
)
def test_empty_file(in_process, hg_repo_path, hg_sha):
    # Add empty file
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    testfile = hg_repo_path / "X"
    testfile.touch()
    hg_out("add")
    hg_out("commit", "-m", "A")
    sha = hg_sha()
    mozphab.main(["submit", "--yes", "--bug", "1", sha], is_development=True)
    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - A

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": None,
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {"unix:filemode": "100644"},
                        "commitHash": mock.ANY,
                        "type": 1,
                        "fileType": 1,
                        "hunks": [],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "hg",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-hg",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "default",
            },
        )
        in call_conduit.call_args_list
    ), (
        "The diff should have populated fields and "
        "`change.hunks` should be an empty list"
    )

    # Modify a file's mode
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-2", diffid="2")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="124")),
        # differential.setdiffproperty
        dict(),
    )
    os.chmod(testfile, 0o0755)

    hg_out("commit", "-m", "B")
    sha = hg_sha()
    mozphab.main(["submit", "--yes", "--bug", "1", sha], is_development=True)
    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - B

Differential Revision: http://example.test/D124
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {"unix:filemode": "100644"},
                        "newProperties": {"unix:filemode": "100755"},
                        "commitHash": mock.ANY,
                        "type": 2,
                        "fileType": 1,
                        "hunks": [],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "hg",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-hg",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "default",
            },
        )
        in call_conduit.call_args_list
    ), (
        "The diff should contain filemode changes and "
        "`change.hunks` should be an empty list"
    )

    # Remove an empty file
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-3", diffid="3")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="125")),
        # differential.setdiffproperty
        dict(),
    )
    testfile.unlink()
    hg_out("addremove")
    hg_out("commit", "-m", "C")
    sha = hg_sha()
    mozphab.main(["submit", "--yes", "--bug", "1", sha], is_development=True)
    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - C

Differential Revision: http://example.test/D125
"""

    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {"unix:filemode": "100755"},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 3,
                        "fileType": 1,
                        "hunks": [],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "hg",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-hg",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "default",
            },
        )
        in call_conduit.call_args_list
    ), (
        "The diff should have empty `newProperties` and "
        "`change.hunks` should be an empty list"
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozphab.logger import logger
from mozphab.environment import MOZPHAB_NAME, MOZPHAB_VERSION


def log_current_version(_):
    logger.info("%s (%s)", MOZPHAB_NAME, MOZPHAB_VERSION)


def add_parser(parser):
    ver_parser = parser.add_parser("version", help="Get version number.")
    ver_parser.set_defaults(func=log_current_version, needs_repo=False)
